import ast
import base64
import io
import os
import re

import pandas as pd
from PIL import Image, ImageChops

from promoai.general_utils import constants

from promoai.model_generation.code_extraction import execute_code_and_get_variable


def image_to_base64(file_pathway: str, max_dim=768) -> str:
    with Image.open(file_pathway) as img:
        # Greyscale and trim whitespaces
        img = img.convert("L")
        bg = Image.new(img.mode, img.size, img.getpixel((0, 0)))
        diff = ImageChops.difference(img, bg)
        bbox = diff.getbbox()
        if bbox:
            img = img.crop(bbox)
        # Ratio-based resizing to minimize token usage
        ratio = max_dim / float(max(img.size))
        if ratio < 1.0:
            new_size = tuple([int(x * ratio) for x in img.size])
            img = img.resize(new_size, Image.Resampling.LANCZOS)

        buffer = io.BytesIO()
        img.save(
            buffer, format="JPEG", quality=60
        )  # Quality 60 is plenty for line graphs
        return base64.b64encode(buffer.getvalue()).decode("utf-8")


def parse_dataframe_for_llms(file_pathway: str) -> str:

    if os.path.getsize(file_pathway) == 0:
        return "The artifact exists but contains no data."

    df = pd.read_csv(file_pathway)
    return df.to_markdown()


def transform_dataframe_for_llms(dataframe: pd.DataFrame) -> str:
    if dataframe.empty:
        return "The artifact exists but contains no data."
    # return it as latex string encapsuled in ```latex ```
    if len(dataframe) < 50:
        return f"```markdown\n{dataframe.to_markdown(index = False)}\n```"
    else:
        statistics_per_column = {
            col: f"{dataframe[col].dtype}, {dataframe[col].nunique()} unique values"
            for col in dataframe.columns
        }
        # add mean, median, std for numeric columns
        numeric_cols = dataframe.select_dtypes(include="number").columns
        for col in numeric_cols:
            statistics_per_column[
                col
            ] += f", mean: {dataframe[col].mean():.2f}, median: {dataframe[col].median():.2f}, std: {dataframe[col].std():.2f}"
        stats_string = "\n".join(
            [f"- {col}: {stats}" for col, stats in statistics_per_column.items()]
        )
        return f"The dataframe has {len(dataframe)} rows and {len(dataframe.columns)} columns. Here are the statistics per column: \n{stats_string}"


def get_expected_event_log_format() -> str:
    return (
        "The expected event log format is a pandas DataFrame with at least the following columns:\n"
        "- 'case:concept:name': Identifier for each case (string)\n"
        "- 'concept:name': Name of the activity (string)\n"
        "- 'time:timestamp': Timestamp of the event (datetime)\n"
        "- 'org:resource': Resource associated with the event (string), this is optional\n\n"
        "Additional columns may be present, but these are the minimum required for process mining tasks."
    )


def _check_whitelisted_libraries(generated_code: str):
    whitelist = {
        "pandas",
        "pm4py",
        "promoai",
        "matplotlib",
        "seaborn",
        "numpy",
        "plotly",
        "datetime",
    }
    try:
        tree = ast.parse(generated_code)
    except SyntaxError as e:
        raise Exception(f"Generated code contains syntax errors: {e}")

    for node in ast.walk(tree):
        # import pandas as pd
        if isinstance(node, ast.Import):
            for alias in node.names:
                top_level = alias.name.split(".")[0]
                if top_level not in whitelist:
                    raise Exception(
                        f"Generated code contains imports from non-whitelisted libraries: {top_level}"
                    )

        # from pm4py.objects.log import x
        elif isinstance(node, ast.ImportFrom):
            if node.module is None:
                raise Exception(
                    "Generated code contains syntax errors: relative imports are not allowed"
                )

            top_level = node.module.split(".")[0]
            if top_level not in whitelist:
                raise Exception(
                    f"Generated code contains imports from non-whitelisted libraries: {top_level}"
                )


def _check_code_for_compilation(generated_code: str):
    try:
        compile(generated_code, "<string>", "exec")
    except SyntaxError as e:
        raise Exception(f"Generated code contains syntax errors: {e}")


def code_extraction_report(code_snippet: str, args, valid_artifact_ids: list):
    """
    Extracts code from a given code snippet, removing any markdown formatting.
    """
    # Check that the code is wrapped in ```python ... ```
    pattern = r"```python\s*(.*?)\s*```"
    match = re.search(pattern, code_snippet, re.DOTALL)

    if not match:
        raise ValueError(
            "Code snippet is not properly formatted with ```python ... ```"
        )

    code = match.group(1).strip()
    if constants.ENABLE_PRINTS:
        print(f"Extracted code:\n{code}")
    _check_whitelisted_libraries(code)
    _check_code_for_compilation(code)

    try:
        final_report = execute_code_and_get_variable(code, "final_report")
    except Exception as e:
        raise Exception(f"Error during code execution: {e}")
    if not isinstance(final_report, list):
        raise Exception(
            "The variable 'final_report' must be a list of dictionaries containing the report text and references to visualizations."
        )
    for entry in final_report:
        if entry["type"] == "artifact" and entry["content"] not in valid_artifact_ids:
            raise Exception(
                f"Invalid artifact reference in final_report: {entry['content']} is not a valid artifact key."
            )
        elif entry["type"] not in ["text", "artifact"]:
            raise Exception(
                f"Invalid type in final_report: {entry['type']}. Expected 'text' or 'artifact'."
            )

    return final_report, None



from typing import Any, List

from powl.objects.obj import POWL

from promoai.general_utils.llm_connection import generate_result_with_error_handling
from promoai.model_generation.code_extraction import (
    execute_code_and_get_variable,
    extract_final_python_code,
)
from promoai.model_generation.validation import (
    validate_partial_orders_with_missing_transitive_edges,
    validate_resource_structure,
)


def extract_model_from_response(
    response: str, auto_duplicate: False
) -> tuple[str, POWL]:
    if auto_duplicate:
        response = response.replace("ModelGenerator()", "ModelGenerator(True, True)")
    extracted_code = extract_final_python_code(response)
    variable_name = "final_model"
    result = execute_code_and_get_variable(extracted_code, variable_name)
    result = result.simplify()
    # validate_unique_transitions(result)
    validate_partial_orders_with_missing_transitive_edges(result)
    validate_resource_structure(result)
    return extracted_code, result


def generate_model(
    conversation: List[dict[str:str]],
    api_key: str,
    llm_name: str,
    ai_provider: str,
    llm_args: dict = None,
    max_iterations=10,
    additional_iterations=5,
) -> tuple[str, POWL, list[Any]]:
    return generate_result_with_error_handling(
        conversation=conversation,
        extraction_function=extract_model_from_response,
        api_key=api_key,
        llm_name=llm_name,
        ai_provider=ai_provider,
        llm_args=llm_args,
        max_iterations=max_iterations,
        additional_iterations=additional_iterations,
    )



"""
    This file includes example (description, model) pairs used for few-shots learning. Some process description are
    from the PET data set "Patrizio Bellan, Han van der Aa, Mauro Dragoni, Chiara Ghidini, and Simone Paolo Ponzetto.
    PET: an annotated dataset for process extraction from natural language text tasks. In Business Process Management
    Workshops 2022, Revised Selected Papers, volume 460 of Lecture Notes in Business Information Processing,
    pages 315–321. Springer, 2022".
"""


from promoai.model_generation import ModelGenerator

d1 = (
    "in this process, you can either do 'a' or 'b'. If 'a' is selected,"
    " then it can be repeated. After completing 'a' or 'b', 'c' may be executed. c is always followed by 'd'."
    " Finally another"
    " execution of 'a' is performed. The whole process is optional and can be skipped."
)


def m1():
    gen = ModelGenerator()
    a = gen.activity("a")
    a_looped = gen.self_loop(a)
    a_copy = gen.copy(a)
    b = gen.activity("b")
    c = gen.activity("c")
    d = gen.activity("d")
    seq_c_d = gen.partial_order(dependencies=[(c, d)])
    skippable_c_d = gen.skip(seq_c_d)
    dg = gen.decision_graph(
        dependencies=[
            (None, a_looped),
            (None, b),
            (a_looped, skippable_c_d),
            (b, skippable_c_d),
            (skippable_c_d, a_copy),
            (a_copy, None),
        ]
    )
    final_model = gen.skip(dg)
    return final_model


def r_m1():
    gen = ModelGenerator()
    a = gen.activity("a", pool=None, lane=None)
    a_copy = gen.copy(a)
    a_looped = gen.self_loop(a)
    b = gen.activity("b", pool=None, lane=None)
    c = gen.activity("c", pool=None, lane=None)
    d = gen.activity("d", pool=None, lane=None)
    seq_c_d = gen.partial_order(dependencies=[(c, d)])
    skippable_c_d = gen.skip(seq_c_d)
    dg = gen.decision_graph(
        dependencies=[
            (None, a_looped),
            (None, b),
            (a_looped, skippable_c_d),
            (b, skippable_c_d),
            (skippable_c_d, a_copy),
            (a_copy, None),
        ]
    )
    final_model = gen.skip(dg)
    return final_model


e1 = (
    "a common error for this process is to add a dependency 'd -> a' without creating a"
    " copy of 'a'. This would imply that the whole process can be executed once again, which is not stated in the description."
    " Another common error is not to mark the sequence c -> d as skippable."
)

r_e1 = (
    e1
    + " Another common error is to assign lanes and pools as there are no roles or organizations mentioned in the description."
)

d1_2 = (
    "in this process, you can either do 'a' or 'b'. If 'a' is selected,"
    " then it can be repeated. After completing 'a' or 'b', 'c' is executed, followed by 'd'. Finally, the process"
    " either ends or goes back to 'a'."
)


def m1_2():
    gen = ModelGenerator()
    a = gen.activity("a")
    a_looped = gen.self_loop(a)
    b = gen.activity("b")
    c = gen.activity("c")
    d = gen.activity("d")
    final_model = gen.decision_graph(
        dependencies=[
            (None, a_looped),
            (None, b),
            (a_looped, c),
            (b, c),
            (c, d),
            (d, a_looped),
            (d, None),
        ]
    )
    return final_model


def r_m1_2():
    gen = ModelGenerator()
    a = gen.activity("a", pool=None, lane=None)
    a_looped = gen.self_loop(a)
    b = gen.activity("b", pool=None, lane=None)
    c = gen.activity("c", pool=None, lane=None)
    d = gen.activity("d", pool=None, lane=None)
    final_model = gen.decision_graph(
        dependencies=[
            (None, a_looped),
            (None, b),
            (a_looped, c),
            (b, c),
            (c, d),
            (d, a_looped),
            (d, None),
        ]
    )
    return final_model


e1_2 = (
    "A common error for this process is to create a copy of 'a' instead of looping back. Another "
    "common mistake here would be to break the process into separable parts, e.g., partial order expressing (c,d)."
    "This is not necessary, as decision graphs can express simple sequential dependencies."
)

r_e1_2 = (
    e1_2
    + " Another common error is to assign lanes and pools as there are no roles or organizations mentioned in the description."
)

d2 = (
    "inventory management can proceed through restocking items or fulfilling orders. Restocking can be performed as "
    "often as necessary. Following either restocking or order fulfillment, an inventory audit is carried out. If"
    " unexpected behavior is detected in the inventory audit, then a data analysis is performed."
    " Additionally, urgent restocking needs can bypass regular restocking and "
    "order fulfillment processes directly leading to the inventory audit. This entire process is modular and can be "
    "repeated or skipped based on operational requirements."
)


def m2():
    gen = ModelGenerator()
    restock = gen.self_loop(gen.activity("restock items"))
    fulfil = gen.activity("fulfill orders")
    urgent_restock = gen.activity("urgent restock")
    inventory_audit = gen.activity("inventory audit")
    data_analysis = gen.activity("data analysis")
    dg = gen.decision_graph(
        dependencies=[
            (None, restock),
            (None, fulfil),
            (None, urgent_restock),
            (restock, inventory_audit),
            (fulfil, inventory_audit),
            (urgent_restock, inventory_audit),
            (inventory_audit, data_analysis),
            (inventory_audit, None),
            (data_analysis, None),
        ]
    )
    final_model = gen.skip(gen.self_loop(dg))
    return final_model


def r_m2():
    gen = ModelGenerator()
    restock = gen.self_loop(gen.activity("restock items", pool=None, lane=None))
    fulfil = gen.activity("fulfill orders", pool=None, lane=None)
    urgent_restock = gen.activity("urgent restock", pool=None, lane=None)
    inventory_audit = gen.activity("inventory audit", pool=None, lane=None)
    data_analysis = gen.activity("data analysis", pool=None, lane=None)
    dg = gen.decision_graph(
        dependencies=[
            (None, restock),
            (None, fulfil),
            (None, urgent_restock),
            (restock, inventory_audit),
            (fulfil, inventory_audit),
            (urgent_restock, inventory_audit),
            (inventory_audit, data_analysis),
            (inventory_audit, None),
            (data_analysis, None),
        ]
    )
    final_model = gen.skip(gen.self_loop(dg))
    return final_model


e2 = (
    "a common error for this process is to copy 'inventory_audit'. Another common error is to create a partial order for"
    " sequential dependencies that can be expressed via the decision graph. Additionally, a common mistake is to omit the self-loop on 'restock items' or on the whole process."
)

r_e2 = (
    e2
    + " Another common error is to assign lanes and pools as there are no roles or organizations mentioned in the description."
)

d3 = (
    "This enhanced payroll process allows for a high degree of customization and adaptation to specific "
    "requirements. Employees' time can be tracked with the option to repeat this step as needed. Pay calculations "
    "follows, incorporating diverse factors such as overtime, bonuses, and deductions. Subsequently, the process "
    "facilitates the issuance of payments and the generation of detailed reports. "
)


def m3():
    gen = ModelGenerator()
    track_time = gen.activity("track time")
    track_time_looped = gen.self_loop(track_time)
    activity_2 = gen.activity("calculate pay")
    activity_3 = gen.activity("issue payments")
    activity_4 = gen.activity("generate reports")
    final_model = gen.partial_order(
        dependencies=[
            (track_time_looped, activity_2),
            (activity_2, activity_3),
            (activity_2, activity_4),
        ]
    )
    return final_model


def r_m3():
    gen = ModelGenerator()
    track_time = gen.activity("track time", pool=None, lane=None)
    track_time_looped = gen.self_loop(track_time)
    activity_2 = gen.activity("calculate pay", pool=None, lane=None)
    activity_3 = gen.activity("issue payments", pool=None, lane=None)
    activity_4 = gen.activity("generate reports", pool=None, lane=None)
    final_model = gen.partial_order(
        dependencies=[
            (track_time_looped, activity_2),
            (activity_2, activity_3),
            (activity_2, activity_4),
        ]
    )
    return final_model


e3 = (
    "a common error for this process is to model a choice between activity_3 and activity_4 instead of the"
    " concurrency. Additionally, a common mistake is to try to model the final model as a decision graph instead"
    " of a partial order. This is not possible here, as decision graphs cannot express the concurrency between"
    " activity_3 and activity_4."
)

r_e3 = (
    e3
    + " Another common error is to assign lanes and pools as there are no roles or organizations mentioned in the description."
)

d4 = (
    "This system combines 4 parallel subprocesses, i.e., that are executed independently/at the same time. The first "
    "process starts with A followed by B then a choice of C and D. The second process consists of a single activity "
    "E which can be repeated but must be executed at least once. The third process consists of the activity F, "
    "which can be repeated or skipped. The last process contains the parallel activities G, H, I, J with the "
    "constrains that I must precede J and H must precede I "
)


def m4():
    gen = ModelGenerator()

    # subprocess 1
    a = gen.activity("a")
    b = gen.activity("b")
    c = gen.activity("c")
    d = gen.activity("d")
    choice_c_d = gen.decision_graph(
        dependencies=[(None, c), (None, d), (c, None), (d, None)]
    )

    # subprocess 2
    e = gen.activity("e")
    unskippable_self_loop_e = gen.self_loop(e)

    # subprocess 3
    f = gen.activity("f")
    skippable_self_loop_f = gen.skip(gen.self_loop(f))

    # subprocess 4
    g = gen.activity("g")
    h = gen.activity("h")
    i = gen.activity("i")
    j = gen.activity("j")

    # combine all subprocesses
    final_model = gen.partial_order(
        dependencies=[
            (a, b),
            (b, choice_c_d),
            (unskippable_self_loop_e,),
            (skippable_self_loop_f,),
            (g,),
            (h, i),
            (i, j),
        ]
    )
    return final_model


def r_m4():
    gen = ModelGenerator()

    # subprocess 1
    a = gen.activity("a", pool=None, lane=None)
    b = gen.activity("b", pool=None, lane=None)
    c = gen.activity("c", pool=None, lane=None)
    d = gen.activity("d", pool=None, lane=None)
    choice_c_d = gen.decision_graph(
        dependencies=[(None, c), (None, d), (c, None), (d, None)]
    )

    # subprocess 2
    e = gen.activity("e", pool=None, lane=None)
    unskippable_self_loop_e = gen.self_loop(e)

    # subprocess 3
    f = gen.activity("f", pool=None, lane=None)
    skippable_self_loop_f = gen.skip(gen.self_loop(f))

    # subprocess 4
    g = gen.activity("g", pool=None, lane=None)
    h = gen.activity("h", pool=None, lane=None)
    i = gen.activity("i", pool=None, lane=None)
    j = gen.activity("j", pool=None, lane=None)

    # combine all subprocesses
    final_model = gen.partial_order(
        dependencies=[
            (a, b),
            (b, choice_c_d),
            (unskippable_self_loop_e,),
            (skippable_self_loop_f,),
            (g,),
            (h, i),
            (i, j),
        ]
    )
    return final_model


e4 = (
    "a common error for this process is to create partial orders for some subprocesses, then trying to add a"
    " partial order as a child of another partial order. Note that the behavior cannot be expressed with one decision graph"
    " because decision graphs cannot express concurrency."
)

r_e4 = (
    e4
    + " Another common error is to assign lanes and pools as there are no roles or organizations mentioned in the description."
)

d5 = (
    "A customer brings in a defective computer and the CRS checks the defect and hands out a repair cost calculation "
    "back. If the customer decides that the costs are acceptable , the process continues , otherwise she takes her "
    "computer home unrepaired. The ongoing repair consists of two activities , which are executed , in an arbitrary "
    "order. The first activity is to check and repair the hardware , whereas the second activity checks and "
    "configures the software. After each of these activities , the proper system functionality is tested. If an "
    "error is detected another arbitrary repair activity is executed , otherwise the repair is finished. "
)


def r_m5():
    gen = ModelGenerator()
    defect_check = gen.activity(
        "Check defect", pool="Repairment Service", lane="Customer Repair Specialist"
    )
    cost_calculation = gen.activity(
        "Calculate repair costs",
        pool="Repairment Service",
        lane="Customer Repair Specialist",
    )
    cancel = gen.activity("Cancel", pool="Customer", lane="Customer")
    return_computer = gen.activity(
        "Give computer unrepaired",
        pool="Repairment Service",
        lane="Customer Repair Specialist",
    )
    repair_hardware = gen.activity(
        "Check and repair the hardware",
        pool="Repairment Service",
        lane="Customer Repair Specialist",
    )
    repair_software = gen.activity(
        "Check and configure the software",
        pool="Repairment Service",
        lane="Customer Repair Specialist",
    )
    test_functionality_after_hardware_repair = gen.activity(
        "Test system functionality",
        pool="Repairment Service",
        lane="Customer Repair Specialist",
    )
    test_functionality_after_software_repair = gen.activity(
        "Test system functionality",
        pool="Repairment Service",
        lane="Customer Repair Specialist",
    )
    additional_hardware_repair = gen.activity(
        "Perform additional hardware repairs",
        pool="Repairment Service",
        lane="Customer Repair Specialist",
    )
    skippable_repair_after_hardware_repair = gen.skip(additional_hardware_repair)
    additional_software_repair = gen.activity(
        "Perform additional software repairs",
        pool="Repairment Service",
        lane="Customer Repair Specialist",
    )
    skippable_repair_after_software_repair = gen.skip(additional_software_repair)
    finish_repair = gen.activity(
        "Finish repair", pool="Repairment Service", lane="Customer Repair Specialist"
    )
    partial_order_repairs = gen.partial_order(
        dependencies=[
            (repair_hardware, test_functionality_after_hardware_repair),
            (repair_software, test_functionality_after_software_repair),
            (
                test_functionality_after_hardware_repair,
                skippable_repair_after_hardware_repair,
            ),
            (
                test_functionality_after_software_repair,
                skippable_repair_after_software_repair,
            ),
        ]
    )
    final_model = gen.decision_graph(
        dependencies=[
            (None, defect_check),
            (defect_check, cost_calculation),
            (cost_calculation, cancel),
            (cancel, return_computer),
            (return_computer, None),
            (cost_calculation, partial_order_repairs),
            (partial_order_repairs, finish_repair),
            (finish_repair, None),
        ]
    )

    return final_model


def m5():
    gen = ModelGenerator()
    defect_check = gen.activity("Check defect")
    cost_calculation = gen.activity("Calculate repair costs")
    cancel = gen.activity("Cancel and give computer unrepaired")
    repair_hardware = gen.activity("Check and repair the hardware")
    repair_software = gen.activity("Check and configure the software")
    test_functionality_after_hardware_repair = gen.activity("Test system functionality")
    test_functionality_after_software_repair = gen.activity("Test system functionality")
    additional_hardware_repair = gen.activity("Perform additional hardware repairs")
    skippable_repair_after_hardware_repair = gen.skip(additional_hardware_repair)
    additional_software_repair = gen.activity("Perform additional software repairs")
    skippable_repair_after_software_repair = gen.skip(additional_software_repair)
    finish_repair = gen.activity("Finish repair")
    partial_order_repairs = gen.partial_order(
        dependencies=[
            (repair_hardware, test_functionality_after_hardware_repair),
            (repair_software, test_functionality_after_software_repair),
            (
                test_functionality_after_hardware_repair,
                skippable_repair_after_hardware_repair,
            ),
            (
                test_functionality_after_software_repair,
                skippable_repair_after_software_repair,
            ),
        ]
    )
    final_model = gen.decision_graph(
        dependencies=[
            (None, defect_check),
            (defect_check, cost_calculation),
            (cost_calculation, cancel),
            (cancel, None),
            (cost_calculation, partial_order_repairs),
            (partial_order_repairs, finish_repair),
            (finish_repair, None),
        ]
    )

    return final_model


e5 = (
    "a common error for this process is to create partial orders for some subprocesses, then trying to add a"
    " partial order as a child of another partial order. Another very important error you should avoid is to"
    " create a local choice between 'cancel' and some local activity (e.g., 'continue process') instead of"
    " modeling a choice between 'cancel' and the rest of the process. Additionally, a common mistake is to integrate the partial order into the decision graph."
)

r_e5 = (
    e5
    + ' Another common error is to create more pools and lanes. Only two pools should be used: "Repairment Service" and "Customer". The "Repairment Service" pool should contain the lane "Customer Repair Specialist", while the "Customer" pool should contain the lane "Customer".'
)

d6 = (
    "A small company manufactures customized bicycles. Whenever the sales department receives an order , "
    "a new process instance is created. A member of the sales department can then reject or accept the order "
    "for a customized bike. In the former case , the process instance is finished. In the latter case , "
    "the storehouse and the engineering department are informed. The storehouse immediately processes the part "
    "list of the order and checks the required quantity of each part. If the part is available in-house , "
    "it is reserved. If it is not available , it is back-ordered. This procedure is repeated for each item on "
    "the part list. In the meantime, the engineering department prepares everything for the assembling of the "
    "ordered bicycle. If the storehouse has successfully reserved or back-ordered every item of the part list "
    "and the preparation activity has finished, the engineering department assembles the bicycle. Afterwards "
    ", the sales department ships the bicycle to the customer and finishes the process instance . "
)


def r_m6():
    gen = ModelGenerator()
    create_process = gen.activity(
        "Create process instance",
        pool="Bike manufacturing Company",
        lane="Sales Department",
    )
    reject_order = gen.activity(
        "Reject order", pool="Bike manufacturing Company", lane="Sales Department"
    )
    accept_order = gen.activity(
        "Accept order", pool="Bike manufacturing Company", lane="Sales Department"
    )
    inform = gen.activity(
        "Inform storehouse and engineering department",
        pool="Bike manufacturing Company",
        lane="Sales Department",
    )
    process_part_list = gen.activity(
        "Process part list", pool="Bike manufacturing Company", lane="Storehouse"
    )
    check_part = gen.activity(
        "Check required quantity of the part",
        pool="Bike manufacturing Company",
        lane="Storehouse",
    )
    reserve = gen.activity(
        "Reserve part", pool="Bike manufacturing Company", lane="Storehouse"
    )
    back_order = gen.activity(
        "Back-order part", pool="Bike manufacturing Company", lane="Storehouse"
    )
    prepare_assembly = gen.activity(
        "Prepare bicycle assembly",
        pool="Bike manufacturing Company",
        lane="Engineering Department",
    )
    assemble_bicycle = gen.activity(
        "Assemble bicycle",
        pool="Bike manufacturing Company",
        lane="Engineering Department",
    )
    ship_bicycle = gen.activity(
        "Ship bicycle", pool="Bike manufacturing Company", lane="Sales Department"
    )
    finish_process = gen.activity(
        "Finish process instance",
        pool="Bike manufacturing Company",
        lane="Sales Department",
    )

    part_subprocess = gen.decision_graph(
        dependencies=[
            (None, check_part),
            (check_part, reserve),
            (check_part, back_order),
            (reserve, None),
            (back_order, None),
        ]
    )

    part_subprocess = gen.self_loop(part_subprocess)
    concurrency = gen.partial_order(
        dependencies=[(part_subprocess,), (prepare_assembly,)]
    )

    final_model = gen.decision_graph(
        dependencies=[
            (None, create_process),
            (create_process, reject_order),
            (create_process, accept_order),
            (reject_order, finish_process),
            (accept_order, inform),
            (inform, process_part_list),
            (process_part_list, concurrency),
            (concurrency, assemble_bicycle),
            (assemble_bicycle, ship_bicycle),
            (ship_bicycle, finish_process),
            (finish_process, None),
        ]
    )

    return final_model


def m6():
    gen = ModelGenerator()
    create_process = gen.activity("Create process instance")
    reject_order = gen.activity("Reject order")
    accept_order = gen.activity("Accept order")
    inform = gen.activity("Inform storehouse and engineering department")
    process_part_list = gen.activity("Process part list")
    check_part = gen.activity("Check required quantity of the part")
    reserve = gen.activity("Reserve part")
    back_order = gen.activity("Back-order part")
    prepare_assembly = gen.activity("Prepare bicycle assembly")
    assemble_bicycle = gen.activity("Assemble bicycle")
    ship_bicycle = gen.activity("Ship bicycle")
    finish_process = gen.activity("Finish process instance")

    part_subprocess = gen.decision_graph(
        dependencies=[
            (None, check_part),
            (check_part, reserve),
            (check_part, back_order),
            (reserve, None),
            (back_order, None),
        ]
    )

    part_subprocess = gen.self_loop(part_subprocess)
    concurrency = gen.partial_order(
        dependencies=[(part_subprocess,), (prepare_assembly,)]
    )

    final_model = gen.decision_graph(
        dependencies=[
            (None, create_process),
            (create_process, reject_order),
            (create_process, accept_order),
            (reject_order, finish_process),
            (accept_order, inform),
            (inform, process_part_list),
            (process_part_list, concurrency),
            (concurrency, assemble_bicycle),
            (assemble_bicycle, ship_bicycle),
            (ship_bicycle, finish_process),
            (finish_process, None),
        ]
    )

    return final_model


e6 = "a common error is not to create a partial order for the concurrency between the part handling subprocess and the preparation of the assembly."
r_e6 = (
    e6
    + 'There is exactly one organization here: "Bike manufacturing Company". There are three lanes: "Sales Department", "Storehouse", and "Engineering Department". Splitting each lane into a different pool is a common mistake.'
)


d7 = (
    "A and B can happen in any order (concurrent). C and D can happen in any order. A precedes both C and D. B "
    "precedes D"
)


def m7():
    gen = ModelGenerator()
    a = gen.activity("A")
    b = gen.activity("B")
    c = gen.activity("C")
    d = gen.activity("D")
    final_model = gen.partial_order(dependencies=[(a, c), (a, d), (b, d)])
    return final_model


def r_m7():
    gen = ModelGenerator()
    a = gen.activity("A", pool=None, lane=None)
    b = gen.activity("B", pool=None, lane=None)
    c = gen.activity("C", pool=None, lane=None)
    d = gen.activity("D", pool=None, lane=None)
    final_model = gen.partial_order(dependencies=[(a, c), (a, d), (b, d)])
    return final_model


e7 = (
    "a common error for this process is to generate a first partial order for modeling the concurrency between"
    " 'A' and 'B', then a second partial order to model the concurrency between 'C' and 'D', then combining these"
    " two partial orders with a large partial that has a sequential dependency from the first order to the second"
    " one. This behavior is not justified and it will imply a wrong dependency ('B' -> 'C'); 'B' and 'C' should"
    " remain independent in the correct partial order."
)
r_e7 = (
    e7
    + " Another common error is to assign lanes and pools as there are no roles or organizations mentioned in the description."
)


d8 = "A followed by B or C. Then D or G. A can be followed by G. Optionally, after A, we can skip all other activities."


def m8():
    gen = ModelGenerator()

    a = gen.activity("A")
    b = gen.activity("B")
    c = gen.activity("C")
    d = gen.activity("D")
    g = gen.activity("G")
    final_model = gen.decision_graph(
        dependencies=[
            (None, a),
            (a, b),
            (a, g),
            (a, c),
            (b, d),
            (b, g),
            (c, g),
            (c, d),
            (d, None),
            (g, None),
            (a, None),
        ]
    )
    return final_model


def r_m8():
    gen = ModelGenerator()

    a = gen.activity("A", pool=None, lane=None)
    b = gen.activity("B", pool=None, lane=None)
    c = gen.activity("C", pool=None, lane=None)
    d = gen.activity("D", pool=None, lane=None)
    g = gen.activity("G", pool=None, lane=None)
    final_model = gen.decision_graph(
        dependencies=[
            (None, a),
            (a, b),
            (a, g),
            (a, c),
            (b, d),
            (b, g),
            (c, g),
            (c, d),
            (d, None),
            (g, None),
            (a, None),
        ]
    )
    return final_model


e8 = "A common error for this process is to include useless partial order structures instead of using one decision graph to model the whole behavior."
r_e8 = (
    e8
    + " Another common error is to assign lanes and pools as there are no roles or organizations mentioned in the description."
)

d9 = (
    "The process starts with checking part stock availability.",
    "After that, we either cancel the order because we don't have the required parts to produce it,"
    "or we continue with the production. When we cancel the order, we notify the customer via e-mail."
    "If we can produce the machines, we check the production schedule and "
    "we schedule the production of each machine. In the end, we produce the machines."
    "As soon as the machines are produced, we ship the machines. In the end, we notify the customer via e-mail or via the system.",
)


def m9():
    gen = ModelGenerator()

    check_part = gen.activity("Check production schedule")
    schedule_part = gen.activity("Schedule production")
    prod_part = gen.activity("Produce machines")
    ship_part = gen.activity("Ship machines")
    production_block = gen.partial_order(
        dependencies=[
            (check_part, schedule_part),
            (schedule_part, prod_part),
            (prod_part, ship_part),
        ]
    )

    check_stock = gen.activity("Check part stock availability")
    cancel = gen.activity("Cancel order")
    notify_email = gen.activity("Notify via email")
    notify_system = gen.activity("Notify via system")

    final_model = gen.decision_graph(
        dependencies=[
            (None, check_stock),
            (check_stock, cancel),
            (check_stock, production_block),
            (cancel, notify_email),
            (production_block, notify_email),
            (production_block, notify_system),
            (notify_email, None),
            (notify_system, None),
        ]
    )

    return final_model


def r_m9():
    gen = ModelGenerator()

    check_part = gen.activity("Check production schedule", pool=None, lane=None)
    schedule_part = gen.activity("Schedule production", pool=None, lane=None)
    prod_part = gen.activity("Produce machines", pool=None, lane=None)
    ship_part = gen.activity("Ship machines", pool=None, lane=None)
    production_block = gen.partial_order(
        dependencies=[
            (check_part, schedule_part),
            (schedule_part, prod_part),
            (prod_part, ship_part),
        ]
    )

    check_stock = gen.activity("Check part stock availability", pool=None, lane=None)
    cancel = gen.activity("Cancel order", pool=None, lane=None)
    notify_email = gen.activity("Notify via email", pool=None, lane=None)
    notify_system = gen.activity("Notify via system", pool=None, lane=None)

    final_model = gen.decision_graph(
        dependencies=[
            (None, check_stock),
            (check_stock, cancel),
            (check_stock, production_block),
            (cancel, notify_email),
            (production_block, notify_email),
            (production_block, notify_system),
            (notify_email, None),
            (notify_system, None),
        ]
    )

    return final_model


e9 = (
    "A common mistake here is to model the decision graph as partial orders of exclusive choices."
    "The sub-model containing the decision graph, i.e., the activities for production, cancelling the order, and the notifications"
    "Is a decision graph, as only one of these paths can be executed in the end: either we start with the production, and then choose between one of the two notificaiton methods, or we cancel the order and notify the customer via e-mail."
    "If we model the decision graph as a partial order of exclusive choices, we lose precision, i.e., the model will allow for more than the described behaviour."
)
r_e9 = (
    e9
    + " Another common error is to assign lanes and pools as there are no roles or organizations mentioned in the description."
)


"""
Machine repairment service, the example is taken from "Activity Instance Identification using Bipartite
Graph Matching", C-Y Li et al.
"""

d10 = (
    "A process for equipment repair and maintenance service is observed."
    "First, a customer sends a machine for repairment (SendMachine)."
    "As soon as the machine is received by the service, the repair is registered (RegisterRepair)."
    "Then,  the technicians analyze the defects (AnalyzeDefects),  while the  customer service department checks the warranty"
    "(CheckWarranty)."
    "Afterwards, if a repair is possible, first, the technicians dismantle the machine (DismantleMachine)"
    "and then repair the faulty parts (RepairPart). If a repair is infeasible,"
    "then the customer service orders a new machine (OrderMachine)."
    "In the end, the repaired or the new machine is shipped back"
    "to the customer by the shipping department (ShipBack)."
)


def r_m10():
    gen = ModelGenerator()
    send_machine = gen.activity("SendMachine", pool="Customer", lane="Customer")
    register_repair = gen.activity(
        "RegisterRepair", pool="RepairService", lane="Customer Service"
    )
    analyze_defects = gen.activity(
        "AnalyzeDefects", pool="RepairService", lane="Technicians"
    )
    check_warranty = gen.activity(
        "CheckWarranty", pool="RepairService", lane="Customer Service"
    )
    dismantle_machine = gen.activity(
        "DismantleMachine", pool="RepairService", lane="Technicians"
    )
    repair_part = gen.activity("RepairPart", pool="RepairService", lane="Technicians")
    repair_part_loop = gen.self_loop(repair_part)
    order_machine = gen.activity(
        "OrderMachine", pool="RepairService", lane="Customer Service"
    )
    ship_back = gen.activity(
        "ShipBack", pool="RepairService", lane="Shipping Department"
    )
    decision_repair_or_order = gen.decision_graph(
        dependencies=[
            (None, dismantle_machine),
            (None, order_machine),
            (dismantle_machine, repair_part_loop),
            (repair_part_loop, None),
            (order_machine, None),
        ]
    )
    partial_order = gen.partial_order(
        dependencies=[
            (send_machine, register_repair),
            (register_repair, analyze_defects),
            (register_repair, check_warranty),
            (analyze_defects, decision_repair_or_order),
            (check_warranty, decision_repair_or_order),
            (decision_repair_or_order, ship_back),
        ]
    )
    return partial_order


r_e10 = (
    "Mind that check warranty and analyze defects can be executed concurrently. This necessities the use of a partial order here."
    "Additionally, there should be a loop on repair part, as multiple parts might need to be repaired."
    'Only two pools should be used: "Repairment Service" and "Customer". The "Repairment Service" pool should contain the lanes "Technicians", "Customer Service", and "Shipping Department", while the "Customer" pool should contain the lane "Customer". Splitting each lane into a different pool is a common mistake.'
)

# d11 = (
#     "This process begins when a customer signs up for a subscription service,"
#     "entering personal and payment information. The system generates an account"
#     ", assigns access, and triggers automated billing cycles."
#     "Throughout the subscription, the customer receives regular updates,"
#     "product enhancements, or renewal notifications. If the customer decides to cancel,"
#     "they submit a cancellation request, which the support team processes. "
#     "Depending on the terms, any refunds or charges are applied."
#     "The process concludes when the subscription is deactivated by the support team"
#     " and the final account balance is settled."
# )
# def r_m11():
#     gen = ModelGenerator()
#     sign_up = gen.activity("Sign up", pool =  "Customer", lane = "Customer")
#     generate_account = gen.activity("Generate account", pool =  "Subscription Service", lane = "System")
#     assign_access = gen.activity("Assign access", pool =  "Subscription Service", lane = "System")
#     trigger_billing = gen.activity("Trigger billing", pool =  "Subscription Service", lane = "System")
#     send_updates = gen.activity("Send updates and notifications", pool =  "Subscription Service", lane = "System")
#     submit_cancellation = gen.activity("Submit cancellation request", pool =  "Customer", lane = "Customer")
#     process_cancellation = gen.activity("Process cancellation", pool =  "Subscription Service", lane = "Support Team")
#     apply_refunds_charges = gen.activity("Apply refunds or charges", pool =  "Subscription Service", lane = "Support Team")
#     deactivate_subscription = gen.activity("Deactivate subscription", pool =  "Subscription Service", lane = "Support Team")
#     settle_final_balance = gen.activity("Settle final account balance", pool =  "Subscription Service", lane = "Support Team")
#
#     billing_loop = gen.self_loop(trigger_billing)
#     update_loop = gen.self_loop(send_updates)
#
#     skip_apply_refunds_charges = gen.skip(apply_refunds_charges)
#
#     cancellation_path = gen.skip(gen.partial_order(
#         dependencies=[
#             (submit_cancellation, process_cancellation),
#             (process_cancellation, skip_apply_refunds_charges),
#             (skip_apply_refunds_charges, settle_final_balance),
#             (process_cancellation, deactivate_subscription),
#         ]
#     ))
#
#     final_model = gen.partial_order(
#         dependencies=[
#             (sign_up, generate_account),
#             (generate_account, assign_access),
#             (assign_access, billing_loop),
#             (assign_access, update_loop),
#             (update_loop, cancellation_path),
#             (billing_loop, cancellation_path),
#         ]
#     )
#
#
#     return final_model
# r_e11 = (
#     'A common mistake is to not include loops for the billing cycle and updates/notifications, as these are recurring activities throughout the subscription period.'
#     'Only two pools should be used: "Subscription Service" and "Customer". The "Subscription Service" pool should contain the lanes "System" and "Support Team", while the "Customer" pool should contain the lane "Customer". Putting each lane into a different pool is a common mistake.'
# )


# d7 = "An employee purchases a product or service he requires. For instance , a sales person on a trip rents a car. " \
#      "The employee submits an expense report with a list of items , along with the receipts for each item. A " \
#      "supervisor reviews the expense report and approves or rejects the report. Since the company has expense rules , " \
#      "there are circumstances where the supervisor can accept or reject the report upon first inspection. These rules " \
#      "could be automated , to reduce the workload on the supervisor. If the supervisor rejects the report , " \
#      "the employee , who submitted it , is given a chance to edit it , for example to correct errors or better " \
#      "describe an expense. If the supervisor approves the report , it goes to the treasurer. The treasurer checks " \
#      "that all the receipts have been submitted and match the items on the list. If all is in order , the treasurer " \
#      "accepts the expenses for processing ( including , e.g., payment or refund , and accounting ). If receipts are " \
#      "missing or do not match the report , he sends it back to the employee. If a report returns to the employee for " \
#      "corrections , it must again go to a supervisor , even if the supervisor previously approved the report. If the " \
#      "treasurer accepts the expenses for processing , the report moves to an automatic activity that links to a " \
#      "payment system. The process waits for the payment confirmation. After the payment is confirmed , the process " \
#      "ends . "
#
#
# def m7():
#     # Base activities
#     purchase = 'Employee purchases product or service'
#     submit_expense = 'Submit expense report with receipts'
#     review_expense_initial = gen.xor('Initially approve report', 'Initially reject report')
#     edit_report = 'Edit expense report'
#     review_expense_after_edit = gen.xor('Approve report after edit', 'Reject report after edit')
#     treasurer_review = gen.xor('Treasurer accepts expenses', 'Treasurer sends back for corrections')
#     automatic_payment = 'Automatic activity for payment system'
#     wait_payment_confirmation = 'Wait for payment confirmation'
#     end_process = 'End process'
#
#     # Subprocess for editing and re-review after rejection at the initial review
#     edit_and_initial_review = gen.partial_order(list_children=[edit_report, review_expense_initial],
#                                     list_order_dependencies=[(edit_report, review_expense_initial)])
#
#     # Subprocess for editing and re-review after the treasurer sends it back for corrections
#     edit_and_treasurer_review = gen.partial_order(list_children=[edit_report, review_expense_after_edit],
#                                       list_order_dependencies=[(edit_report, review_expense_after_edit)])
#
#     # Combining treasurer review decision with possible loop back for corrections
#     treasurer_decision = DoRedoLoop('Treasurer accepts expenses', edit_and_treasurer_review)
#
#     # Main process flow
#     main_process_flow = gen.partial_order(
#         list_children=[submit_expense, review_expense_initial, treasurer_decision, automatic_payment,
#                        wait_payment_confirmation, end_process],
#         list_order_dependencies=[(submit_expense, review_expense_initial),
#                                  (review_expense_initial, treasurer_decision),
#                                  (treasurer_decision, automatic_payment),
#                                  (automatic_payment, wait_payment_confirmation),
#                                  (wait_payment_confirmation, end_process)])
#
#     # Final model starts with the purchase and follows the main process flow
#     final_model = gen.partial_order(list_children=[purchase, main_process_flow],
#                         list_order_dependencies=[(purchase, main_process_flow)])
#
#     return final_model


SHOTS = [
    (d1, m1, e1),
    (d1_2, m1_2, e1_2),
    (d2, m2, e2),
    (d3, m3, e3),
    (d4, m4, e4),
    (d5, m5, e5),
    (d6, m6, e6),
    (d7, m7, e7),
    (d8, m8, e8),
    (d9, m9, e9),
]
RESOURCE_AWARE_SHOTS = [
    (d1, r_m1, r_e1),
    (d1_2, r_m1_2, r_e1_2),
    (d2, r_m2, r_e2),
    (d3, r_m3, r_e3),
    (d4, r_m4, r_e4),
    (d5, r_m5, r_e5),
    (d6, r_m6, r_e6),
    (d7, r_m7, r_e7),
    (d8, r_m8, r_e8),
    (d9, r_m9, r_e9),
    (d10, r_m10, r_e10),
    # (d11, r_m11, r_e11),
]

if __name__ == "__main__":
    from pm4py import view_petri_net
    from powl import convert_to_petri_net, view

    model = m9().simplify()
    view(model)
    pn, im, fm = convert_to_petri_net(model)
    view_petri_net(pn, im, fm, format="SVG")



from promoai.prompting.prompt_engineering import cut_conversation


def test_cut_history_1():
    conversation = [
        {
            "role": "user",
            "content": 'Your role: you are an expert in process modeling, familiar with common process constructs such as exclusive choice, do-redo loops, and partial orders. Your task is to analyze the textual description of a process and transform it into a process model in the POWL language. When generating a model, be as precise as possible and capture all details of the process in the model. Also act as the process owner and use your expertise and familiarity with the process context to fill in any missing knowledge. \n\nAdd…ed.Only two pools should be used: "Repairment Service" and "Customer". The "Repairment Service" pool should contain the lanes "Technicians", "Customer Service", and "Shipping Department", while the "Customer" pool should contain the lane "Customer". Splitting each lane into a different pool is a common mistake.\n\nAt the end of your response provide a single Python code snippet (i.e., staring with \'```python\') that contains the full final code. \n\nThis is the process description: A then B',
        },
        {
            "role": "assistant",
            "content": '```python\nfrom promoai.model_generation.generator import ModelGenerator\ngen = ModelGenerator()\na = gen.activity("A", pool = None, lane = None)\nb = gen.activity("B", pool = None, lane = None)\nfinal_model = gen.partial_order(dependencies=[(a, b)])\n```',
        },
        {
            "role": "user",
            "content": "Please update the model to fix it based on the provided feedback. Please make sure the returned model matches the initial process description, all previously provided feedback, and the newfeedback comment as well. Make sure to save the updated final model is the variable 'final_model'. This is the new feedback comment:add C",
        },
        {
            "role": "assistant",
            "content": '```python\nfrom promoai.model_generation.generator import ModelGenerator\ngen = ModelGenerator()\na = gen.activity("A", pool = None, lane = None)\nb = gen.activity("B", pool = None, lane = None)\nc = gen.activity("C", pool = None, lane = None)\nfinal_model = gen.partial_order(dependencies=[(a, b), (b, c)])\n```',
        },
        {
            "role": "user",
            "content": "Please update the model to fix it based on the provided feedback. Please make sure the returned model matches the initial process description, all previously provided feedback, and the newfeedback comment as well. Make sure to save the updated final model is the variable 'final_model'. This is the new feedback comment:add new pool and D in it",
        },
        {
            "role": "assistant",
            "content": '```python\nfrom promoai.model_generation.generator import ModelGenerator\ngen = ModelGenerator()\na = gen.activity("A", pool = None, lane = None)\nb = gen.activity("B", pool = None, lane = None)\nc = gen.activity("C", pool = None, lane = None)\nd = gen.activity("D", pool = "New Pool", lane = None)\nfinal_model = gen.partial_order(dependencies=[(a, b), (b, c), (c, d)])\n```',
        },
    ]
    index = 4
    cut_conv = cut_conversation(conversation, index)
    assert cut_conv[-1]["content"] == conversation[4]["content"]


def test_cut_history():
    conversation = [
        {"role": "user", "content": "You are a helpful assistant."},
        {"role": "user", "content": "Hello!"},
        {
            "role": "user",
            "content": "Please update the model to fix it based on the provided feedback.",
            "type": "error",
        },
        {"role": "assistant", "content": "Hi there! How can I help you?"},
        {"role": "user", "content": "Can you tell me a joke?"},
        {
            "role": "assistant",
            "content": "Why did the scarecrow win an award? Because he was outstanding in his field!",
        },
        {"role": "user", "content": "That's funny! Tell me another one."},
        {
            "role": "user",
            "content": "Please update the model to fix it based on the provided feedback.",
        },
    ]
    index = 6
    cut_conv = cut_conversation(conversation, index)
    assert cut_conv[-1]["content"] == "That's funny! Tell me another one."
    index = -1
    # We expect an exception
    try:
        cut_conv = cut_conversation(conversation, index)
        assert False, "Expected an exception for negative index"
    except ValueError:
        pass
    index = 10
    # We expect an exception
    try:
        cut_conv = cut_conversation(conversation, index)
        assert False, "Expected an exception for out-of-bounds index"
    except ValueError:
        pass
    index = 4
    cut_conv = cut_conversation(conversation, index)
    assert cut_conv[-1]["content"] == "Can you tell me a joke?"
    index = 3
    cut_conv = cut_conversation(conversation, index)
    assert cut_conv[-1]["content"] == "Hello!"



import copy
from functools import partial
from typing import Tuple

from promoai.agents.pm4py_wrapper import PM4PYWrapper
from promoai.agents.state import ProcessState
from promoai.agents.utils import code_extraction_report
from promoai.general_utils.llm_connection import (
    generate_result_with_error_handling,
    LLMConnection,
)


def init_state(user_request: str, event_log) -> ProcessState:
    initial_state = ProcessState(user_request=user_request, event_log=event_log)
    return initial_state


def engineer_node(
    state: ProcessState, LLMCredentials: LLMConnection
) -> Tuple[ProcessState, str, str]:
    # Initial state
    # 1. Construct Prompt
    api = PM4PYWrapper(state)
    api_summary = api.get_API_summary()

    msg = f"""
    You are a Process Mining Data Engineer.

    Your task is to preprocess the event log data according to the user's request, using the provided API methods and any necessary data manipulation with pandas or numpy.
    {api_summary} \n

    DATA: \n
    - Assume that the log is already loaded and accessible via api.event_log. \n
    - Furthermore, the data has the following structure: {state['log_abstraction']} \n
    - Activities are in the 'concept:name' column, timestamps in 'time:timestamp', case identifier in 'case:concept:name'. \n

    EXPECTED OUTPUT:
    - Always return the final event log as `final_event_log` variable after any preprocessing steps, so it can be passed down to the subsequent agent.

    IMPORTANT:
    - If the previous code you generated is able to answer the user's new request without any modifications, just return the event log without any changes.


    """
    msg_history = (
        [
            {"role": "system", "content": msg},
            {"role": "user", "content": state["user_request"][-1]},
        ]
        if len(state["messages_eng"]) == 0
        else state["messages_eng"]
        + [{"role": "user", "content": state["user_request"][-1]}]
    )
    code, result, messages = generate_result_with_error_handling(
        msg_history,
        extraction_function=api.code_extraction,
        llm_name=LLMCredentials.llm_name,
        ai_provider=LLMCredentials.ai_provider,
        api_key=LLMCredentials.api_key,
        llm_args=LLMCredentials.args,
        max_iterations=3,
        additional_iterations=2,
    )
    state["event_log"] = result
    state["log_abstraction"] = state.generate_log_abstraction()
    state["messages_eng"] = messages
    return state, result, code


def update_message_for_analyst(last_request: str, context: str) -> str:
    msg = f"""
    The Process Engineer has executed the following preprocessing steps on the event log based on the user's request "{context}":
    Please use this information to generate a report, making sure to reference artifacts that are relevant to the current user request and to support your analysis with the provided visualizations and data summaries.

    REMEMBER:
    - YOUR GOAL is to answer the User Request: "{last_request}" by presenting and interpreting the provided artifacts. \n
    - Recall the output format: \n
    ```python
    final_report = [
        {{"type": "text", "content": "### Analysis of Process Distribution\\nAs requested, I have analyzed the activity distribution..."}},
        {{"type": "artifact", "content": "artifact_0"}},
        {{"type": "text", "content": "The chart above (artifact_0) clearly shows that..."}}
    ]
    ```
    \n
    - Make sure to follow the strict rules introduced in the initial message (e.g., referencing artifacts by their exact keys, including the visualization if the user asked for it and it is provided, etc.) when generating your report.
    - Do NOT mention any errors or internal processing steps. Your report should be a polished analysis that directly addresses the user's request using the provided artifacts as evidence to support your conclusions.
    You will receive the artifacts in the consequent message.
    """

    return msg


def generate_initial_message_for_analyst(state: ProcessState, context) -> str:
    msg = f"""
    You are a process analyst.

    CRITICAL INSTRUCTION:
    The Process Engineer has already executed the analysis. Your SOLE TASK is to compile the final report using the artifacts provided below.
    - DO NOT ask the user if they want to generate a visualization; it has ALREADY been generated.
    - DO NOT say "I can generate..." or "If you want me to...".
    - YOUR GOAL is to answer the User Request: "{state['user_request'][-1]}" by presenting and interpreting the provided artifacts.

    DATA SUMMARY:
    - Log Abstraction: {state['log_abstraction']}
    - Additional Context: {context}

    TASK:
    Construct the `final_report` variable. Interleave your analysis with the relevant artifact keys (e.g., "artifact_0").
    - If an artifact is a visualization (PNG), use it to support your findings.
    - If an artifact is a table (CSV), summarize the key insights in text rather than displaying the whole table instead of adding the entire potentially large table to your report unless explicitly asked by the user.
    - For visualizations, you will get an abstraction in form of a description (content of the artifact or its summary). Use it to identify if it is relevant and to support your analysis.

    OUTPUT FORMAT:
    Your output must be Python code that defines a variable `final_report` wrapped in a python code block (```python ... ```).

    Example:
    ```python
    final_report = [
        {{"type": "text", "content": "### Analysis of Process Distribution\\nAs requested, I have analyzed the activity distribution..."}},
        {{"type": "artifact", "content": "artifact_0"}},
        {{"type": "text", "content": "The chart above (artifact_0) clearly shows that..."}}
    ]
    ```

    STRICT RULES:
    1. Reference artifacts by their EXACT keys (e.g., "artifact_0").
    2. Do NOT mention any errors or internal processing steps.
    3. If the user asked for a visualization and it is provided in the ARTIFACTS section, you MUST include it in the report.
    4. You may reuse artifacts from previous iterations by referencing their keys, but you cannot introduce new artifacts that were not provided in the ARTIFACTS section.
    5. Before generating the report, carefully analyze the provided artifacts, respond after making sure you have fully utilized the available information to answer the user's request.
    """
    return msg


def analyst_node(state: ProcessState, LLMCredentials: LLMConnection) -> ProcessState:
    artifact_id_to_description_and_content = {}
    artifact_id_to_filepath = {}
    sent_artifacts = state["sent_artifacts"]

    for i, (file_path, (description, data)) in enumerate(
        state["saved_artifacts"].items()
    ):
        aid = f"artifact_{i}"
        clean_description = str(description).strip("{}'\" ")
        artifact_id_to_filepath[aid] = file_path

        a_type = (
            "visualization"
            if file_path.endswith((".png", ".jpg", ".jpeg"))
            else "dataframe"
            if file_path.endswith(".csv")
            else None
        )

        if a_type is None:
            raise Exception(
                f"Unsupported artifact type for artifact {file_path}. Only image and csv files are supported."
            )

        a_content = (
            "Visual representation available (PNG). Reference this key to display the chart"
            if a_type == "visualization"
            else data
        )

        artifact_id_to_description_and_content[aid] = {
            "description": clean_description,
            "content": a_content,
            "type": a_type,
        }

    artifact_dataframes = {
        k: v
        for k, v in artifact_id_to_description_and_content.items()
        if v["type"] == "dataframe"
    }
    artifact_df_strings = "\n".join(
        [
            f"\n =====DATAFRAME {k}======:\n Description: {v['description']} \n Content: {v['content']} \n \
                                ======END OF {k} ====== \n\n"
            for k, v in artifact_dataframes.items()
            if k not in sent_artifacts
        ]
    )

    artifact_viz = {
        k: v
        for k, v in artifact_id_to_description_and_content.items()
        if v["type"] == "visualization"
    }
    artifact_viz_strings = "\n".join(
        [
            f"\n =====VISUALIZATION {k}======:\n Description: {v['description']} \n Content: {v['content']} \n \n ======END OF {k} ====== \n\n"
            for k, v in artifact_viz.items()
            if k not in sent_artifacts
        ]
    )
    context = (
        " ".join(state["context"])
        if len(state["context"]) > 0
        else "No additional context available."
    )
    msg_history = state["messages_ana"]
    additional_info = f"These are the artifacts that have been generated in the current iteration \
        and are relevant to the user's request but have not been included in the previous report: \n \n \
        {artifact_df_strings} \n\n  \
        {artifact_viz_strings} \n\n \
        Recall the user request: {state['user_request'][-1]} \n\n "

    if len(msg_history) == 0:
        initial_msg = generate_initial_message_for_analyst(state, context)
        msg_history.append({"role": "system", "content": initial_msg})
    else:
        updated_msg = update_message_for_analyst(state["user_request"][-1], context)
        msg_history.append({"role": "user", "content": updated_msg})

    msg_history.append({"role": "user", "content": additional_info})

    valid_artifact_ids = list(artifact_id_to_description_and_content.keys())
    partial_function = partial(
        code_extraction_report, valid_artifact_ids=valid_artifact_ids
    )
    try:
        result, _, messages = generate_result_with_error_handling(
            msg_history,
            extraction_function=partial_function,
            llm_name=LLMCredentials.llm_name,
            ai_provider=LLMCredentials.ai_provider,
            api_key=LLMCredentials.api_key,
            llm_args=LLMCredentials.args,
            max_iterations=3,
            additional_iterations=2,
        )
        state["sent_artifacts"].extend(
            list(artifact_id_to_description_and_content.keys())
        )

    except Exception as e:
        raise Exception(f"Error during analyst node execution: {e}")

    # Now, we have to postprocess the result to replace the image keys with their actual content (e.g., file paths)
    postprocessed_report = copy.deepcopy(result)
    for entry in postprocessed_report:
        if entry.get("type") == "artifact":
            key = entry.get("content")
            if key in artifact_id_to_filepath:
                entry["content"] = artifact_id_to_filepath[key]
            else:
                raise ValueError(f"Key {key} not found.")
    state["messages_ana"] = messages
    state["final_report"] = postprocessed_report
    state.flush_context()
    return state



import xml.etree.ElementTree as ET
from io import StringIO


def get_simplified_xml_abstraction(bpmn_content):
    root = ET.fromstring(bpmn_content)
    output = StringIO()

    def display_element(element, indent=0):

        indent_str = " " * indent
        elem_type = element.tag.split("}")[-1]
        elem_id = element.get("id")

        label = None
        if element.get("name"):
            elm_name = " ".join(element.get("name").strip().split())
            if len(elm_name) > 0:
                label = elm_name

        if element.text:
            elm_text = " ".join(element.text.strip().split())
            if len(elm_text) > 0:
                if label:
                    label = label + " | " + elm_text
                else:
                    label = elm_text

                    # leaf_substrings = ['event', 'flow', 'task', 'gateway', 'association']
        attrs_to_exclude = [
            "type",
            "id",
            "targetNamespace",
            "name",
            "exporter",
            "exporterVersion",
            "isExecutable",
        ]
        children_to_exclude = ["incoming", "outgoing"]

        # def is_leaf(element_type):
        #     return any(sub in element_type.lower() for sub in leaf_substrings)

        attr_keys = [
            key for key in element.attrib.keys() if key not in attrs_to_exclude
        ]
        children = [
            child
            for child in element
            if child.tag.split("}")[-1] not in children_to_exclude
            and child.tag.startswith("{http://www.omg.org/spec/BPMN/20100524/MODEL}")
        ]

        # if len(attr_keys) > 0 or (not is_leaf(elem_type) and list(element)):
        if len(attr_keys) > 0 or len(children) > 0:
            # Hierarchical representation
            id = " " + elem_id if elem_id else ""
            name = f" ({label})" if label else ""
            output.write(f"{indent_str}<{elem_type}{id}>" + name)
            output.write("\n")
            for attr, value in element.attrib.items():
                if attr not in attrs_to_exclude:
                    output.write(f"{indent_str}  - {attr}: {value}")
                    output.write("\n")
            for child in children:
                display_element(child, indent + 4)

            output.write(f"{indent_str}</{elem_type}>")
            output.write("\n")
        else:
            # Flat representation
            id = " " + elem_id if elem_id else ""
            name = f" ({label})" if label else ""
            output.write(f"{indent_str}<{elem_type}{id}" + name + "/>")
            output.write("\n")

    # Begin processing from the root element
    display_element(root)
    return output.getvalue()



from typing import Dict, List

from promoai.model_generation.code_extraction import (
    execute_code_and_get_variable,
    extract_final_python_code,
)


def extraction_function_dictionary(response: str, keys: List[str]) -> tuple[str, Dict]:
    extracted_code = extract_final_python_code(response)
    print(extracted_code)
    variable_name = "score_dictionary"
    score_dict = execute_code_and_get_variable(extracted_code, variable_name)
    if not isinstance(score_dict, Dict):
        raise Exception(f"Not a dictionary! Found: {score_dict}!")
    if list(score_dict.keys()) == keys:
        return extracted_code, score_dict
    else:
        raise Exception(
            f"Wrong keys in the extracted dictionary! Expected: {keys}; found: {score_dict.keys()}!"
        )


def generate_self_evaluation_prompt(
    description: str, model_codes: Dict[str, str], conformance_evaluation: bool = False
) -> str:
    """
    Generates a prompt for the LLM to evaluate multiple models based on a process description.

    Args:
        description (str): The process description.
        model_codes (Dict[str, str]): A dictionary where keys are model IDs (e.g., 'IT1') and values are
         the corresponding POWL codes.
        conformance_evaluation

    Returns:
        str: The formatted prompt to be sent to the LLM.
    """
    prompt = """You have already generated 4 candidate POWL models for a single process descriptions
     following the previously detailed task. Now, let's forget about the process modeling role and take the role of
     an assistant specialized in evaluating the already generated process models.
     Your task is to assess the quality of the provided POWL models based on how well they capture the behavior
      described in the process description.\n\n"""
    prompt = prompt + f"""This was the process description: {description}\n\n"""
    prompt = (
        prompt
        + f"""The generated POWL models are saved in the following dictionary:: {model_codes}\n\n"""
    )

    if conformance_evaluation:
        prompt += f"""
        **Evaluation Criteria:**
        - **Fitness:** Evaluate how well the process model can reproduce the behaviors of the process according to the
         process description.
        - **Precision:** Evaluate the extent to which the process model exclusively represents behaviors that
         are allowed in the process according to the process description. """
    else:
        prompt += f"""
        **Evaluation Criteria:**
        - **Behavior Accuracy:** How accurately does the model capture the intended process behavior?
        - **Completeness:** Does the model include all necessary activities as described?
        - **Correctness:** Are the control flows (e.g., partial orders, choices, loops) correctly implemented?"""

    prompt += f"""**Output Requirements:**
    - Provide your evaluation scores as a Python dictionary named `score_dictionary`.
    - Use the same keys for your dictionary as the keys in the provided dictionary, i.e., {list(model_codes.keys())}."""

    prompt += """
    - Each value should be a float between 0 and 1, where:
        - **1** indicates the model perfectly captures the described behavior.
        - **0** indicates the model does not capture the described behavior at all.

    **Example Format:**
    ```python
    score_dictionary = {
        'IT1': 0.85,
        'IT2': 0.90,
        'IT3': 0.80,
        'IT4': 0.95
    }
    ```
    Please ensure to include at the end of your response a Python snipped with a dictionary strictly
     following the above Python code format. """

    return prompt



from promoai.aipa.conversation import (
    create_conversation,
    create_message,
    create_process_model_representation,
)
from promoai.general_utils.llm_connection import query_llm


class BPMNAnalyzer:
    def __init__(
        self,
        api_key: str,
        ai_model: str,
        ai_provider: str,
        initial_query: str,
        bpmn_xml_string: str = None,
        bpmn_json_string: str = None,
        selected_elements_json: str = None,
        model_abstraction="simplified_xml",
        enable_role_prompting=True,
        enable_natural_language_restriction=True,
        enable_chain_of_thought=False,
        enable_process_analysis=False,
        enable_knowledge_injection=False,
        enable_few_shots_learning=False,
        enable_negative_prompting=False,
        enable_examples=True,
    ):

        self.api_key = api_key
        self.ai_model = ai_model
        self.ai_provider = ai_provider
        self.model_abstraction = model_abstraction
        self._conversation = create_conversation(
            role="system",
            model_abstraction=model_abstraction,
            enable_role_prompting=enable_role_prompting,
            enable_natural_language_restriction=enable_natural_language_restriction,
            enable_chain_of_thought=enable_chain_of_thought,
            enable_process_analysis=enable_process_analysis,
            enable_knowledge_injection=enable_knowledge_injection,
            enable_few_shots_learning=enable_few_shots_learning,
            enable_negative_prompting=enable_negative_prompting,
            enable_examples=enable_examples,
        )

        self._conversation.append(
            create_process_model_representation(
                model_abstraction=model_abstraction,
                role="system",
                xml_string=bpmn_xml_string,
                json_abstraction=bpmn_json_string,
                selected_elements_json=selected_elements_json,
            )
        )

        self.last_response = None
        self.ask(initial_query)

    def get_last_response(self) -> str | None:
        return self.last_response

    def ask(self, query: str) -> str | None:

        user_message_content = create_message(
            query, role="user", model_abstraction=self.model_abstraction
        )
        self._conversation.append(user_message_content)

        self.last_response = query_llm(
            conversation=self._conversation,
            api_key=self.api_key,
            llm_name=self.ai_model,
            ai_provider=self.ai_provider,
        )
        self._conversation.append({"role": "assistant", "content": self.last_response})



import pm4py
from pm4py.util import constants
from powl import convert_to_bpmn, convert_to_petri_net, view as view_powl

from powl.objects.obj import POWL

from promoai.model_generation import code_extraction
from promoai.model_generation.model_generation import generate_model

from promoai.pn_to_powl.converter_utils.powl_to_code import translate_powl_to_code
from promoai.prompting import create_conversation, update_conversation


class LLMProcessModelGenerator:
    def __init__(self, process_model, conversation):
        self.process_model = process_model
        self.conversation = conversation

    @classmethod
    def from_description(
        cls,
        process_description: str,
        api_key: str,
        ai_model: str,
        ai_provider: str,
        resource_aware_discovery: bool = False,
        llm_args: dict = None,
    ):
        init_conversation = create_conversation(
            process_description, resource_aware_discovery=resource_aware_discovery
        )
        code, process_model, conversation = generate_model(
            init_conversation,
            api_key=api_key,
            llm_name=ai_model,
            ai_provider=ai_provider,
            llm_args=llm_args,
        )
        return cls(process_model, conversation)

    @classmethod
    def from_powl(cls, powl_model: POWL):
        init_conversation = create_conversation(None, resource_aware_discovery=False)
        conversation = list(init_conversation)
        conversation.append(
            {
                "role": "user",
                "content": "Instead of starting with a process description, I will give the code of"
                " the initial process model:\n\n" + translate_powl_to_code(powl_model),
            }
        )
        return cls(powl_model, conversation)

    def get_conversation(self):
        return self.conversation

    def get_code(self):
        return code_extraction.extract_final_python_code(
            self.conversation[-1]["content"]
        )

    def get_powl(self):
        return self.process_model

    def get_petri_net(self):
        return convert_to_petri_net(self.process_model)

    def get_bpmn(self):
        bpmn_model = convert_to_bpmn(self.process_model)
        return bpmn_model

    def update(
        self,
        feedback: str,
        api_key: str,
        ai_model: str,
        ai_provider: str,
        llm_args: dict = None,
    ):
        self.conversation = update_conversation(self.conversation, feedback)
        code, self.process_model, self.conversation = generate_model(
            conversation=self.conversation,
            api_key=api_key,
            llm_name=ai_model,
            ai_provider=ai_provider,
            llm_args=llm_args,
        )
        self.process_model = self.process_model.simplify()

    def view_bpmn(self, image_format: str = "svg"):
        bpmn_model = self.get_bpmn()
        pm4py.view_bpmn(bpmn_model, format=image_format)

    def view_petri_net(self, image_format: str = "svg"):
        net, im, fm = self.get_petri_net()
        pm4py.view_petri_net(net, im, fm, format=image_format)

    def view_powl(self, image_format: str = "svg"):
        view_powl(self.process_model)

    def export_bpmn(self, file_path: str, encoding: str = constants.DEFAULT_ENCODING):
        if not file_path.lower().endswith("bpmn"):
            raise Exception(
                "The provided file path does not have the '.bpmn' extension!"
            )
        bpmn_model = self.get_bpmn()
        from pm4py.objects.bpmn.exporter import exporter

        exporter.apply(bpmn_model, file_path, parameters={"encoding": encoding})

    def export_petri_net(
        self, file_path: str, encoding: str = constants.DEFAULT_ENCODING
    ):
        if not file_path.lower().endswith("pnml"):
            raise Exception(
                "The provided file path does not have the '.pnml' extension!"
            )
        net, im, fm = self.get_petri_net()
        from pm4py.objects.petri_net.exporter import exporter as petri_exporter

        petri_exporter.apply(
            net=net,
            initial_marking=im,
            final_marking=fm,
            output_filename=file_path,
            parameters={"encoding": encoding},
        )


# def initialize(process_description: str | None, api_key: str, llm_name: str, ai_provider: str,
#                powl_model: POWL = None, n_candidates: int = 1, debug: bool = False):
#     best_cand = None
#     exception = ""
#     for i in range(n_candidates):
#         try:
#             cand = LLMProcessModelGenerator(process_description=process_description, api_key=api_key,
#                                             llm_name=llm_name, ai_provider=ai_provider,
#                                             powl_model=powl_model)
#             if n_candidates > 1:
#                 raise Exception("Currently, there is no support for multiple candidate generation!")
#             else:
#                 best_cand = cand
#         except Exception as e:
#             exception = str(e)
#     if best_cand is None:
#         raise Exception(exception)
#     return best_cand
#
#
# def update(generator: LLMProcessModelGenerator, feedback: str, api_key: str, llm_name: str,
#            ai_provider: str, n_candidates: int = 1, debug: bool = False):
#     best_cand = None
#     exception = ""
#     for i in range(n_candidates):
#         try:
#             cand = generator if n_candidates == 1 else deepcopy(generator)
#             cand.update(feedback, api_key, llm_name, ai_provider)
#             if n_candidates > 1:
#                 raise Exception("Currently, there is no support for multiple candidate generation!")
#             else:
#                 best_cand = cand
#         except Exception as e:
#             exception = str(e)
#     if best_cand is None:
#         raise Exception(exception)
#     return best_cand



from pm4py.objects.process_tree.obj import Operator
from powl.objects.BinaryRelation import BinaryRelation
from powl.objects.obj import (
    DecisionGraph,
    EndNode,
    OperatorPOWL,
    SilentTransition,
    StartNode,
    StrictPartialOrder,
    Transition,
)

from promoai.prompting.prompt_engineering import import_statement


def translate_powl_to_code(powl_obj):
    """
    Translates a POWL object from pm4py into code using ModelGenerator.

    Args:
        powl_obj: The POWL object to translate.

    Returns:
        A string containing the Python code that constructs the equivalent POWL model using ModelGenerator.
    """
    code_lines = [import_statement, "gen = ModelGenerator()"]

    var_counter = [0]

    def get_new_var_name():
        var_name = f"var_{var_counter[0]}"
        var_counter[0] += 1
        return var_name

    def process_powl(powl):
        if isinstance(powl, Transition):
            var_name = get_new_var_name()
            if isinstance(powl, SilentTransition):
                code_lines.append(f"{var_name} = None")
            else:
                label = powl.label
                if powl._organization is not None and powl._role is not None:
                    code_lines.append(
                        f"{var_name} = gen.activity('{label}', pool='{powl._organization}',  lane='{powl._role}')"
                    )
                else:
                    code_lines.append(f"{var_name} = gen.activity('{label}')")
            return var_name

        elif isinstance(powl, StartNode) or isinstance(powl, EndNode):
            return None

        elif isinstance(powl, OperatorPOWL):
            operator = powl.operator
            children = powl.children
            if operator == Operator.XOR:
                rel = BinaryRelation(children)
                graph = DecisionGraph(rel, children, children, False)
                graph = graph.reduce_silent_transitions()
                return process_powl(graph)
            elif operator == Operator.LOOP:
                rel = BinaryRelation(children)
                do = children[0]
                redo = children[1]
                rel.add_edge(do, redo)
                rel.add_edge(redo, do)
                graph = DecisionGraph(rel, [do], [do], False)
                graph = graph.reduce_silent_transitions()
                return process_powl(graph)
            else:
                raise Exception("Unknown operator! This should not be possible!")

        elif isinstance(powl, StrictPartialOrder) or isinstance(powl, DecisionGraph):
            nodes = powl.order.nodes
            if isinstance(powl, StrictPartialOrder):
                order = powl.order.get_transitive_reduction()
            elif isinstance(powl, DecisionGraph):
                order = powl.order
            else:
                raise Exception("Unknown POWL object! This should not be possible!")
            node_var_map = {node: process_powl(node) for node in nodes}
            dependencies = []
            nodes_in_edges = set()
            for source in nodes:
                for target in nodes:
                    source_var = node_var_map[source]
                    target_var = node_var_map[target]
                    if order.is_edge(source, target):
                        dependencies.append(f"({source_var}, {target_var})")
                        nodes_in_edges.update([source, target])

            if isinstance(powl, StrictPartialOrder):
                # Include nodes not in any edge as singleton tuples
                for node in nodes:
                    if node not in nodes_in_edges:
                        var = node_var_map[node]
                        dependencies.append(f"({var},)")

            dep_str = ", ".join(dependencies)
            var_name = get_new_var_name()
            code_lines.append(
                f"{var_name} = gen.partial_order(dependencies=[{dep_str}])"
            ) if isinstance(powl, StrictPartialOrder) else code_lines.append(
                f"{var_name} = gen.decision_graph(dependencies=[{dep_str}])"
            )
            return var_name

        else:
            raise Exception(
                f"Unknown POWL object {type(powl)}! This should not be possible!"
            )

    final_var = process_powl(powl_obj)
    code_lines.append(f"final_model = {final_var}")

    return "\n".join(code_lines)



from pm4py import PetriNet
from powl import convert_from_workflow_net
from powl.objects.obj import POWL

def convert_workflow_net_to_powl(net: PetriNet) -> POWL:
    """
    Convert a Petri net to a POWL model.

    Parameters:
    - net: PetriNet

    Returns:
    - POWL model
    """
    return convert_from_workflow_net(net)



def add_prompt_strategies(
    model_abstraction="simplified_xml",
    enable_role_prompting=True,
    enable_natural_language_restriction=True,
    enable_chain_of_thought=False,
    enable_process_analysis=False,
    enable_knowledge_injection=False,
    enable_few_shots_learning=False,
    enable_negative_prompting=False,
    enable_examples=True,
):

    prompt = ""

    if enable_role_prompting:
        role_prompt = role_prompting()
        prompt += role_prompt

    if enable_process_analysis:
        proc_anal = role_prompting_process_expert()
        prompt += proc_anal

    if enable_natural_language_restriction:
        nlp_restriction = natural_language_restriction()
        prompt += nlp_restriction

    if enable_chain_of_thought:
        ch_of_thought = chain_of_thoughts()
        prompt += ch_of_thought

    if enable_knowledge_injection:
        know_inj = knowledge_injection()
        prompt += know_inj

    if enable_examples:
        examples = get_examples()
        prompt += examples
    elif model_abstraction in ["json", "simplified_xml"]:
        if enable_few_shots_learning:
            if model_abstraction == "json":
                abstraction = process_textual_representation_json
            elif model_abstraction == "simplified_xml":
                abstraction = process_textual_representation_simplified_xml
            few_shots = few_shots_learning_json(abstraction)
            prompt += few_shots

            if enable_negative_prompting:
                negative_prompt = negative_prompting_json()
                prompt += negative_prompt

        elif enable_negative_prompting:
            if model_abstraction == "json":
                abstraction = process_textual_representation_json
            elif model_abstraction == "simplified_xml":
                abstraction = process_textual_representation_simplified_xml
            negative_prompt = negative_prompting_with_questions(abstraction)
            prompt += negative_prompt

    return prompt


def role_prompting():
    return "- Your role: You are an expert in business process modeling and the BPMN 2.0 standard. I will give you a textual representation of a full BPMN model (and also potentially a selection of a subset of elements) and ask you questions about the process. You are supposed to answer the questions based on your understanding of the provided model.\n\n"


# from https://en.wikipedia.org/wiki/Business_Process_Model_and_Notation
def knowledge_injection():
    return """
- Business Process Model and Notation (BPMN) is a standardized graphical notation for modeling business processes.\n
- BPMN's four basic element categories are:\n
1. Flow objects: Events, activities, gateways\n
2. Connecting objects: Sequence flow, message flow, association\n
3. Swim lanes: Pool, lane, Dark Pool\n
4. Artifacts: Data object, group, annotation\n
- Event: An Event is represented with a circle and denotes something that happens (compared with an activity, which is something that is done). Icons within the circle denote the type of event (e.g., an envelope representing a message, or a clock representing time). Events are also classified as Catching (for example, if catching an incoming message starts a process) or Throwing (such as throwing a completion message when a process ends).\n
1. Start event: Acts as a process trigger; indicated by a single narrow border, and can only be Catch, so is shown with an open (outline) icon.\n
2. Intermediate event: Represents something that happens between the start and end events; is indicated by a double border, and can Throw or Catch (using solid or open icons as appropriate). For example, a task could flow to an event that throws a message across to another pool, where a subsequent event waits to catch the response before continuing.\n
3. End event: Represents the result of a process; indicated by a single thick or bold border, and can only Throw, so is shown with a solid icon.\n
- Activity: An activity is represented with a rounded-corner rectangle and describes the kind of work which must be done. An activity is a generic term for work that a company performs. It can be atomic or compound.\n
1. Task: A task represents a single unit of work that is not or cannot be broken down to a further level of business process detail. It is referred to as an atomic activity. A task is the lowest level activity illustrated on a process diagram. A set of tasks may represent a high-level procedure.\n
2. Sub-process: Used to hide or reveal additional levels of business process detail. When collapsed, a sub-process is indicated by a plus sign against the bottom line of the rectangle; when expanded, the rounded rectangle expands to show all flow objects, connecting objects, and artifacts. A sub-process is referred to as a compound activity. A sub-process has its own self-contained start and end events; sequence flows from the parent process must not cross the boundary.\n
3. Transaction: A form of sub-process in which all contained activities must be treated as a whole; i.e., they must all be completed to meet an objective, and if any one of them fails, they must all be compensated (undone). Transactions are differentiated from expanded sub-processes by being surrounded by a double border.\n
4. Call Activity: A point in the process where a global process or a global Task is reused. A call activity is differentiated from other activity types by a bolded border around the activity area.\n
- Gateway: A gateway is represented with a diamond shape and determines forking and merging of paths, depending on the conditions expressed.\n
1. Exclusive: Used to create alternative flows in a process. Because only one of the paths can be taken, it is called exclusive.\n
2. Event Based: The condition determining the path of a process is based on an evaluated event.\n
3. Parallel: Used to create parallel paths without evaluating any conditions.\n
4. Inclusive: Used to create alternative flows where all paths are evaluated.\n
5. Exclusive Event Based: An event is being evaluated to determine which of mutually exclusive paths will be taken.\n
6. Complex: Used to model complex synchronization behavior.\n
7. Parallel Event Based: Two parallel processes are started based on an event, but there is no evaluation of the event.\n
- Connections: Flow objects are connected to each other using Connecting objects, which are of three types: sequences, messages, and associations.\n
1. Sequence Flow: A Sequence Flow is represented with a solid line and arrowhead, and shows in which order the activities are performed. The sequence flow may also have a symbol at its start, a small diamond indicates one of a number of conditional flows from an activity, while a diagonal slash indicates the default flow from a decision or activity with conditional flows.\n
2. Message Flow: A Message Flow is represented with a dashed line, an open circle at the start, and an open arrowhead at the end. It tells us what messages flow across organizational boundaries (i.e., between pools). A message flow can never be used to connect activities or events within the same pool.\n
3. Association: An Association is represented with a dotted line. It is used to associate an Artifact or text to a Flow Object, and can indicate some directionality using an open arrowhead (toward the artifact to represent a result, from the artifact to represent an input, and both to indicate it is read and updated). No directionality is used when the Artifact or text is associated with a sequence or message flow (as that flow already shows the direction).\n
- Swim lanes are a visual mechanism of organising and categorising activities, based on cross functional flowcharting, and in BPMN consist of two types:\n
- Pool: Represents major participants in a process, typically separating different organisations. A pool contains one or more lanes (like a real swimming pool). A pool can be open (i.e., showing internal detail) when it is depicted as a large rectangle showing one or more lanes, or collapsed (i.e., hiding internal detail) when it is depicted as an empty rectangle stretching the width or height of the diagram.\n
- Lane: Used to organise and categorise activities within a pool according to function or role, and depicted as a rectangle stretching the width or height of the pool. A lane contains the flow objects, connecting objects and artifacts.\n
- Artifacts: Artifacts allow developers to bring some more information into the model/diagram. In this way the model/diagram becomes more readable. There are three pre-defined Artifacts, and they are:\n
1. Data objects: Data objects show the reader which data is required or produced in an activity.\n
2. Group: A Group is represented with a rounded-corner rectangle and dashed lines. The group is used to group different activities but does not affect the flow in the diagram.\n
3. Annotation: An annotation is used to give the reader of the model/diagram an understandable impression.\n\n\n"""


def natural_language_restriction():
    return (
        "- Please answer in natural language, so that any user not familiar with the BPMN standard can understand your answer without any technical knowledge; i.e., avoid IDs or technical terms like Task, Gate, flow, lane, etc.; rather use natural language to describe the behavior of the underlying process.\n\n"
        " Please give short, compact answers.\n\n"
        " Please use the same language as the one used by the user in the last question question.\n\n"
    )


def chain_of_thoughts():
    return "- Where possible (especially for complex queries), please share the chain of thoughts or reasoning behind your answers. This helps in understanding how you arrived at your conclusion.\n\n"


def role_prompting_process_expert():
    return "- Please take the role of a process expert who is familiar with the domain of the provided process, and use your domain knowledge to better understand and analyze the process, filling in any missing gaps.\n\n"


# for the model credit-scoring-asynchronous.bpmn from https://github.com/camunda/bpmn-for-research
process_textual_representation_json = """- { $type: bpmn:Collaboration, id: Collaboration_1y0blh3, participants: [object Object],[object Object],[object Object], messageFlows: [object Object],[object Object],[object Object],[object Object],[object Object],[object Object],[object Object], $parent: Definitions_1 }
- { $type: bpmn:Participant, id: Participant_1x9zkso, name: credit scoring frontend (bank), processRef: Process_1, $parent: Collaboration_1y0blh3 }
- { $type: bpmn:Participant, id: Participant_0e81yis, name: credit scoring (bank), processRef: Process_0hiditg, $parent: Collaboration_1y0blh3 }
- { $type: bpmn:Task, id: Task_16winvj, name: request credit score, $parent: Process_0hiditg }
- { $type: bpmn:ExclusiveGateway, id: ExclusiveGateway_11dldcm, $parent: Process_0hiditg }
- { $type: bpmn:Task, id: Task_1fzfxey, name: send credit score, $parent: Process_0hiditg }
- { $type: bpmn:EndEvent, id: EndEvent_0rp5trg, name: scoring request handled, eventDefinitions: , $parent: Process_0hiditg }
- { $type: bpmn:EndEvent, id: EndEvent_0rp5trg, name: scoring request handled, eventDefinitions: , $parent: Process_0hiditg }
- { $type: bpmn:Task, id: Task_0l942o9, name: report delay, $parent: Process_0hiditg }
- { $type: bpmn:IntermediateCatchEvent, id: IntermediateCatchEvent_0yg7cuh, name: credit score received, eventDefinitions: [object Object], $parent: Process_0hiditg }
- { $type: bpmn:IntermediateCatchEvent, id: IntermediateCatchEvent_0yg7cuh, name: credit score received, eventDefinitions: [object Object], $parent: Process_0hiditg }
- { $type: bpmn:IntermediateCatchEvent, id: IntermediateCatchEvent_0a8iz14, name: credit score received, eventDefinitions: [object Object], $parent: Process_0hiditg }
- { $type: bpmn:IntermediateCatchEvent, id: IntermediateCatchEvent_0a8iz14, name: credit score received, eventDefinitions: [object Object], $parent: Process_0hiditg }
- { $type: bpmn:StartEvent, id: StartEvent_1els7eb, name: scoring request received, eventDefinitions: [object Object], $parent: Process_0hiditg }
- { $type: bpmn:StartEvent, id: StartEvent_1els7eb, name: scoring request received, eventDefinitions: [object Object], $parent: Process_0hiditg }
- { $type: bpmn:IntermediateCatchEvent, id: IntermediateCatchEvent_0ujob24, name: delay information received, eventDefinitions: [object Object], $parent: Process_0hiditg }
- { $type: bpmn:IntermediateCatchEvent, id: IntermediateCatchEvent_0ujob24, name: delay information received, eventDefinitions: [object Object], $parent: Process_0hiditg }
- { $type: bpmn:EventBasedGateway, id: EventBasedGateway_02s95tm, $parent: Process_0hiditg }
- { $type: bpmn:Participant, id: Participant_1xfb3ml, name: scoring service, processRef: Process_1dc1p3b, $parent: Collaboration_1y0blh3 }
- { $type: bpmn:Task, id: Task_07vbn2i, name: send credit score, $parent: Process_1dc1p3b }
- { $type: bpmn:Task, id: Task_01ouvha, name: report delay, $parent: Process_1dc1p3b }
- { $type: bpmn:ExclusiveGateway, id: ExclusiveGateway_0rtdod4, name: score available?, $parent: Process_1dc1p3b }
- { $type: bpmn:ExclusiveGateway, id: ExclusiveGateway_0rtdod4, name: score available?, $parent: Process_1dc1p3b }
- { $type: bpmn:Task, id: Task_06dqs9t, name: send credit score, $parent: Process_1dc1p3b }
- { $type: bpmn:EndEvent, id: EndEvent_0khk0tq, name: scoring request handled, eventDefinitions: , $parent: Process_1dc1p3b }
- { $type: bpmn:EndEvent, id: EndEvent_0khk0tq, name: scoring request handled, eventDefinitions: , $parent: Process_1dc1p3b }
- { $type: bpmn:ExclusiveGateway, id: ExclusiveGateway_125lzox, $parent: Process_1dc1p3b }
- { $type: bpmn:Task, id: Task_02m68xj, name: compute credit score (level 2), $parent: Process_1dc1p3b }
- { $type: bpmn:Task, id: Task_1r15hqs, name: compute credit score (level 1), $parent: Process_1dc1p3b }
- { $type: bpmn:StartEvent, id: StartEvent_0o849un, name: scoring request received, eventDefinitions: [object Object], $parent: Process_1dc1p3b }
- { $type: bpmn:StartEvent, id: StartEvent_0o849un, name: scoring request received, eventDefinitions: [object Object], $parent: Process_1dc1p3b }
- { $type: bpmn:TextAnnotation, id: TextAnnotation_0si9ilm, text: inkl. ID for message\nqueueing, $parent: Process_1dc1p3b }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_0rrtx7k, sourceRef: StartEvent_1els7eb, targetRef: Task_16winvj, $parent: Process_0hiditg }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_1i1amgb, sourceRef: IntermediateCatchEvent_0a8iz14, targetRef: ExclusiveGateway_11dldcm, $parent: Process_0hiditg }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_1fy80l7, sourceRef: IntermediateCatchEvent_0yg7cuh, targetRef: ExclusiveGateway_11dldcm, $parent: Process_0hiditg }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_12a77en, sourceRef: ExclusiveGateway_11dldcm, targetRef: Task_1fzfxey, $parent: Process_0hiditg }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_1nyeozm, sourceRef: Task_1fzfxey, targetRef: EndEvent_0rp5trg, $parent: Process_0hiditg }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_0rf5cxd, sourceRef: IntermediateCatchEvent_0ujob24, targetRef: Task_0l942o9, $parent: Process_0hiditg }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_08fsgff, sourceRef: Task_0l942o9, targetRef: IntermediateCatchEvent_0a8iz14, $parent: Process_0hiditg }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_0o5t8lw, sourceRef: Task_16winvj, targetRef: EventBasedGateway_02s95tm, $parent: Process_0hiditg }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_0e97dad, sourceRef: EventBasedGateway_02s95tm, targetRef: IntermediateCatchEvent_0ujob24, $parent: Process_0hiditg }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_1kdut76, sourceRef: EventBasedGateway_02s95tm, targetRef: IntermediateCatchEvent_0yg7cuh, $parent: Process_0hiditg }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_0jh32vv, name: no, sourceRef: ExclusiveGateway_0rtdod4, targetRef: Task_01ouvha, $parent: Process_1dc1p3b }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_0jh32vv, name: no, sourceRef: ExclusiveGateway_0rtdod4, targetRef: Task_01ouvha, $parent: Process_1dc1p3b }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_052bcer, name: yes, sourceRef: ExclusiveGateway_0rtdod4, targetRef: Task_07vbn2i, $parent: Process_1dc1p3b }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_052bcer, name: yes, sourceRef: ExclusiveGateway_0rtdod4, targetRef: Task_07vbn2i, $parent: Process_1dc1p3b }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_0t0wbx3, sourceRef: ExclusiveGateway_125lzox, targetRef: EndEvent_0khk0tq, $parent: Process_1dc1p3b }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_0dkbeo7, sourceRef: Task_06dqs9t, targetRef: ExclusiveGateway_125lzox, $parent: Process_1dc1p3b }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_1xqy47o, sourceRef: Task_07vbn2i, targetRef: ExclusiveGateway_125lzox, $parent: Process_1dc1p3b }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_08jl5se, sourceRef: Task_02m68xj, targetRef: Task_06dqs9t, $parent: Process_1dc1p3b }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_1yiajt6, sourceRef: Task_01ouvha, targetRef: Task_02m68xj, $parent: Process_1dc1p3b }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_1nznlgx, sourceRef: Task_1r15hqs, targetRef: ExclusiveGateway_0rtdod4, $parent: Process_1dc1p3b }
- { $type: bpmn:SequenceFlow, id: SequenceFlow_158pur5, sourceRef: StartEvent_0o849un, targetRef: Task_1r15hqs, $parent: Process_1dc1p3b }
- { $type: bpmn:Association, id: Association_1ctd4ma, sourceRef: Task_01ouvha, targetRef: TextAnnotation_0si9ilm, $parent: Process_1dc1p3b }
- { $type: bpmn:MessageFlow, id: MessageFlow_1pkfls0, sourceRef: Participant_1x9zkso, targetRef: StartEvent_1els7eb, $parent: Collaboration_1y0blh3 }
- { $type: bpmn:MessageFlow, id: MessageFlow_1m6362g, sourceRef: Task_0l942o9, targetRef: Participant_1x9zkso, $parent: Collaboration_1y0blh3 }
- { $type: bpmn:MessageFlow, id: MessageFlow_1i21wes, sourceRef: Task_1fzfxey, targetRef: Participant_1x9zkso, $parent: Collaboration_1y0blh3 }
- { $type: bpmn:MessageFlow, id: MessageFlow_1mm30jd, sourceRef: Task_16winvj, targetRef: StartEvent_0o849un, $parent: Collaboration_1y0blh3 }
- { $type: bpmn:MessageFlow, id: MessageFlow_1136yi9, sourceRef: Task_07vbn2i, targetRef: IntermediateCatchEvent_0yg7cuh, $parent: Collaboration_1y0blh3 }
- { $type: bpmn:MessageFlow, id: MessageFlow_0bgkr12, sourceRef: Task_06dqs9t, targetRef: IntermediateCatchEvent_0a8iz14, $parent: Collaboration_1y0blh3 }
- { $type: bpmn:MessageFlow, id: MessageFlow_1nwyn8k, sourceRef: Task_01ouvha, targetRef: IntermediateCatchEvent_0ujob24, $parent: Collaboration_1y0blh3 }\n"""

process_textual_representation_simplified_xml = """<definitions Definitions_1>\n    <collaboration Collaboration_1y0blh3>\n        <participant Participant_1x9zkso> (credit scoring frontend (bank))\n          - processRef: Process_1\n
  </participant>\n        <participant Participant_0e81yis> (credit scoring (bank))\n          - processRef: Process_0hiditg\n        </participant>\n        <participant Participant_1xfb3ml> (scoring service)\n          - processRef: Process_1dc1p3b\n        </participant>\n        <messageFlow MessageFlow_1pkfls0>\n          - sourceRef: Participant_1x9zkso\n
   - targetRef: StartEvent_1els7eb\n        </messageFlow>\n        <messageFlow MessageFlow_1m6362g>\n          - sourceRef: Task_0l942o9\n          - targetRef: Participant_1x9zkso\n
       </messageFlow>\n        <messageFlow MessageFlow_1i21wes>\n          - sourceRef: Task_1fzfxey\n          - targetRef: Participant_1x9zkso\n        </messageFlow>\n        <messageFlow MessageFlow_1mm30jd>\n          - sourceRef: Task_16winvj\n          - targetRef: StartEvent_0o849un\n        </messageFlow>\n        <messageFlow MessageFlow_1136yi9>\n
 - sourceRef: Task_07vbn2i\n          - targetRef: IntermediateCatchEvent_0yg7cuh\n        </messageFlow>\n        <messageFlow MessageFlow_0bgkr12>\n          - sourceRef: Task_06dqs9t\n          - targetRef: IntermediateCatchEvent_0a8iz14\n        </messageFlow>\n        <messageFlow MessageFlow_1nwyn8k>\n          - sourceRef: Task_01ouvha\n          - targetRef: IntermediateCatchEvent_0ujob24\n        </messageFlow>\n    </collaboration>\n    <process Process_1>\n      - isExecutable: false\n    </process>\n    <process Process_0hiditg>\n
 <task Task_16winvj (request credit score)/>\n        <exclusiveGateway ExclusiveGateway_11dldcm/>\n        <task Task_1fzfxey (send credit score)/>\n        <endEvent EndEvent_0rp5trg
(scoring request handled)/>\n        <task Task_0l942o9 (report delay)/>\n        <sequenceFlow SequenceFlow_0rrtx7k>\n          - sourceRef: StartEvent_1els7eb\n          - targetRef:
Task_16winvj\n        </sequenceFlow>\n        <sequenceFlow SequenceFlow_1i1amgb>\n          - sourceRef: IntermediateCatchEvent_0a8iz14\n          - targetRef: ExclusiveGateway_11dldcm\n        </sequenceFlow>\n        <sequenceFlow SequenceFlow_1fy80l7>\n          - sourceRef: IntermediateCatchEvent_0yg7cuh\n          - targetRef: ExclusiveGateway_11dldcm\n
</sequenceFlow>\n        <sequenceFlow SequenceFlow_12a77en>\n          - sourceRef: ExclusiveGateway_11dldcm\n          - targetRef: Task_1fzfxey\n        </sequenceFlow>\n        <sequenceFlow SequenceFlow_1nyeozm>\n          - sourceRef: Task_1fzfxey\n          - targetRef: EndEvent_0rp5trg\n        </sequenceFlow>\n        <sequenceFlow SequenceFlow_0rf5cxd>\n
      - sourceRef: IntermediateCatchEvent_0ujob24\n          - targetRef: Task_0l942o9\n        </sequenceFlow>\n        <sequenceFlow SequenceFlow_08fsgff>\n          - sourceRef: Task_0l942o9\n          - targetRef: IntermediateCatchEvent_0a8iz14\n        </sequenceFlow>\n        <intermediateCatchEvent IntermediateCatchEvent_0yg7cuh> (credit score received)\n
      <messageEventDefinition/>\n        </intermediateCatchEvent>\n        <intermediateCatchEvent IntermediateCatchEvent_0a8iz14> (credit score received)\n            <messageEventDefinition/>\n        </intermediateCatchEvent>\n        <startEvent StartEvent_1els7eb> (scoring request received)\n            <messageEventDefinition/>\n        </startEvent>\n        <intermediateCatchEvent IntermediateCatchEvent_0ujob24> (delay information received)\n            <messageEventDefinition/>\n        </intermediateCatchEvent>\n        <sequenceFlow SequenceFlow_0o5t8lw>\n          - sourceRef: Task_16winvj\n          - targetRef: EventBasedGateway_02s95tm\n        </sequenceFlow>\n        <eventBasedGateway EventBasedGateway_02s95tm/>\n        <sequenceFlow SequenceFlow_0e97dad>\n          - sourceRef: EventBasedGateway_02s95tm\n          - targetRef: IntermediateCatchEvent_0ujob24\n        </sequenceFlow>\n
<sequenceFlow SequenceFlow_1kdut76>\n          - sourceRef: EventBasedGateway_02s95tm\n          - targetRef: IntermediateCatchEvent_0yg7cuh\n        </sequenceFlow>\n    </process>\n
  <process Process_1dc1p3b>\n        <sequenceFlow SequenceFlow_0jh32vv> (no)\n          - sourceRef: ExclusiveGateway_0rtdod4\n          - targetRef: Task_01ouvha\n        </sequenceFlow>\n        <sequenceFlow SequenceFlow_052bcer> (yes)\n          - sourceRef: ExclusiveGateway_0rtdod4\n          - targetRef: Task_07vbn2i\n        </sequenceFlow>\n        <sequenceFlow SequenceFlow_0t0wbx3>\n          - sourceRef: ExclusiveGateway_125lzox\n          - targetRef: EndEvent_0khk0tq\n        </sequenceFlow>\n        <sequenceFlow SequenceFlow_0dkbeo7>\n          - sourceRef: Task_06dqs9t\n          - targetRef: ExclusiveGateway_125lzox\n        </sequenceFlow>\n        <sequenceFlow SequenceFlow_1xqy47o>\n          - sourceRef: Task_07vbn2i\n          - targetRef: ExclusiveGateway_125lzox\n        </sequenceFlow>\n        <sequenceFlow SequenceFlow_08jl5se>\n          - sourceRef: Task_02m68xj\n          - targetRef: Task_06dqs9t\n        </sequenceFlow>\n        <sequenceFlow SequenceFlow_1yiajt6>\n          - sourceRef: Task_01ouvha\n          - targetRef: Task_02m68xj\n        </sequenceFlow>\n        <sequenceFlow SequenceFlow_1nznlgx>\n          - sourceRef: Task_1r15hqs\n          - targetRef: ExclusiveGateway_0rtdod4\n        </sequenceFlow>\n        <sequenceFlow SequenceFlow_158pur5>\n          - sourceRef: StartEvent_0o849un\n          - targetRef: Task_1r15hqs\n        </sequenceFlow>\n        <task Task_07vbn2i (send credit score)/>\n        <task Task_01ouvha (report delay)/>\n        <exclusiveGateway ExclusiveGateway_0rtdod4 (score available?)/>\n        <task Task_06dqs9t (send credit score)/>\n        <endEvent EndEvent_0khk0tq (scoring request handled)/>\n        <exclusiveGateway ExclusiveGateway_125lzox/>\n        <task Task_02m68xj (compute credit score (level 2))/>\n        <task Task_1r15hqs (compute credit score (level 1))/>\n        <startEvent StartEvent_0o849un> (scoring request received)\n            <messageEventDefinition/>\n        </startEvent>\n        <association Association_1ctd4ma>\n          - sourceRef: Task_01ouvha\n          - targetRef: TextAnnotation_0si9ilm\n        </association>\n        <textAnnotation TextAnnotation_0si9ilm>\n
  <text (inkl. ID for message queueing)/>\n        </textAnnotation>\n    </process>\n</definitions>\n"""

examples_json = [
    {
        "input": "How does the bank start the credit scoring process?",
        "output_good": "A bank clerk uses their software to request a credit score for a customer, which kicks off the process in the banking system to communicate with the credit protection agency.",
        "output_bad": "The bank initiates a BPMN Collaboration, invoking a Participant Process via a MessageFlow to the Schufa service for a credit scoring Event.",
    },
    {
        "input": "What happens after the bank sends a scoring request to the agency?",
        "output_good": "The agency performs an initial credit scoring, and if this initial scoring is successful, the result is sent back immediately to the bank's system and shown to the clerk.",
        "output_bad": "Upon receipt of the scoring request, an IntermediateCatchEvent is triggered, leading to an immediate execution of a Task for level 1 scoring, conditioned by an ExclusiveGateway.",
    },
    {
        "input": "What occurs if the initial credit scoring doesn’t give an immediate result?",
        "output_good": "The agency informs the bank's system about the delay and starts a more detailed scoring. The bank's system then sends a notification to the clerk to report the delay.",
        "output_bad": "If the initial scoring doesn't resolve, a conditional ExclusiveGateway routes the flow to a delay notification Task, subsequently initializing a level 2 scoring Sub-Process.",
    },
    {
        "input": "How is the clerk informed of the final credit scoring result?",
        "output_good": "Once the scoring is completed and the result is sent back to the bank's system, it is sent to the clerk's software.",
        "output_bad": "The final scoring outcome, after processing through a sequence of EventBasedGateways and Task executions, is programmatically dispatched to the initiating Participant's frontend interface.",
    },
    {
        "input": "What does the bank do if the credit score is delayed?",
        "output_good": "The bank's system sends a notification to the clerk to report the delay, and as soon as the final score is ready, it sends the results to the clerk.",
        "output_bad": "In the event of a scoring delay, an IntermediateCatchEvent captures the timeout, activating a Task to update the UI component of the clerk's software module with a delay notification.",
    },
]


def few_shots_learning_json(abstraction):
    res = "- Let us consider the following textual representation of an example process:\n"
    res += abstraction + "\n\n"
    res += "- These are example pairs of input and expected output:\n"
    for i in range(len(examples_json)):
        res += f"Input {i}: {examples_json[i]['input']}\n"
        res += f"Expected output {i}: {examples_json[i]['output_good']}\n"
    res += "\n\n"
    return res


def negative_prompting_json():
    res = "- Now, I will give you example bad outputs for the same example questions, so you try to avoid such outputs:\n"
    for i in range(len(examples_json)):
        res += f"Bad output for question {i}: {examples_json[i]['output_bad']}\n"
    res += "\n\n"
    return res


def negative_prompting_with_questions(abstraction):
    res = "- Let us consider the following textual representation of an example process:\n"
    res += abstraction + "\n\n"
    res = "- Now, I will give you example pairs of input questions and bad outputs that you should try to avoid:\n"
    for i in range(len(examples_json)):
        res += f"Input {i}: {examples_json[i]['input']}\n"
        res += f"Bad output for question {i}: {examples_json[i]['output_bad']}\n"
    res += "\n\n"
    return res


def get_examples():
    res = (
        "- Now, I will give you example triples of input questions, good answers,"
        " and bad answers that you should try to avoid:\n"
    )
    for i in range(len(examples_json)):
        res += f"Input {i}: {examples_json[i]['input']}\n"
        res += f"Good answer for question {i}: {examples_json[i]['output_good']}\n"
        res += f"Bad answer for question {i}: {examples_json[i]['output_bad']}\n"
    res += "\n\n"
    return res



import inspect
from typing import List, Optional

from promoai.prompting.shots import RESOURCE_AWARE_SHOTS, SHOTS

import_statement = "from promoai.model_generation.generator import ModelGenerator"

ERROR_MESSAGE_FOR_MODEL_GENERATION = """
Please update the model to fix the error. Make sure" \
                          f" to save the updated final model is the variable 'final_model'. """

STRICT_PROMPT = False


def add_role():
    res = (
        "Your role: you are an expert in process modeling,"
        " familiar with common"
        " process constructs such as exclusive choice, do-redo loops, and partial orders."
        " Your task is to analyze the textual description of a process and transform it into a process model in"
        " the POWL language. When generating a model, be as precise"
        " as possible and capture all details of the process in the model. "
    )
    if STRICT_PROMPT:
        res = (
            res
            + "Please create the process model strictly depending on the provided description, without using "
            "any domain knowledge you might have. You are not supposed to correct any information in the "
            "process, rather fully reply on the provided textual description.\n\n"
        )
    else:
        res = (
            res + "Also act as the process owner and use"
            " your expertise and familiarity with the"
            " process context to fill in any missing knowledge. \n\n"
        )
    return res


def add_knowledge(resource_aware_discovery=False):
    prompt = (
        "Use the following knowledge about the POWL process modeling language:\n"
        "A POWL model is a hierarchical model. POWL models are recursively generated"
        " by combining submodels into a new model either using a partial order"
        " or a decision graph. "
        " We define three types of POWL models. The first type of POWL models is the base case consisting of a"
        " single activity. For the second type"
        " of POWL models, we use a decision graph to combine"
        " multiple POWL models into a new model. A decision graph can model complex control flow, where exactly one path is taken at a time."
        " They can model decision patterns: cyclic behavior and exclusive choice structures."
        " The third type of POWL models is defined as a partial"
        " order over n >= 2 submodels. A partial order is binary relation that is irreflexive, transitive,"
        " and asymmetry. "
        "Provide the Python code that "
        "recursively generate a POWL model. Save the final model is the"
        " variable 'final_model'. Do not try to execute the code, just return it. Assume the class ModelGenerator"
        " is properly implemented and can be imported using the import statement:"
        f" {import_statement}. ModelGenerator provides the functions"
        " described below:\n"
    )
    prompt = (
        (
            prompt
            + " - activity(label) generates an activity. It takes 1 string arguments,"
            " which is the label of the activity.\n"
        )
        if not resource_aware_discovery
        else (
            prompt
            + " - activity(label, pool, lane) generates an activity. It takes 3 string arguments,"
            " which are the label of the activity, the pool name, and the lane name."
        )
    )
    prompt = (
        prompt
        + " - partial_order(dependencies) takes 1 argument, which is a list of tuples of submodels. These tuples"
        " set the nodes of the partial order and specify the"
        " edges of the partial order (i.e., the sequential dependencies). The"
        " transitive closure of the added dependencies should conform with the irreflexivity"
        " requirement of partial orders. We interpret unconnected nodes in a partial order to be"
        " concurrent and connections between nodes as sequential dependencies. Use a partial order"
        " with no edges (with the parameter 'dependencies' set to a list of tuples of size 1) to model pure"
        " concurrency/independency; i.e., to model the relation "
        " between sub models that can all be happens at the same time/in any order. However, note that all of them"
        " need to happen unlike the xor case in decision graphs. The main difference is that with xor case you model alternative"
        " paths (either path_1 or path_2), while with a partial order you model concurrent paths (you do both"
        " path_1 and path_2). The general assumption is partial orders is that nodes are concurrent; however, you can"
        " still add sequential dependencies between certain nodes (as tuples in the list for the parameter"
        " 'dependencies'). For example, this is"
        " the case in systems where you execute all subprocesses but one of them must be completed before"
        " starting another one. Assume we have 4 submodel A, B, C, D. partial_order(dependencies=[(A, B), (B, C), (C, D)]) "
        "models a sequence A -> B -> C -> D; partial_order(dependencies=[(A,), (B,), (C,), (D,)]) models full"
        " concurrency; partial_order(dependencies=[(A,B), (C,), (D,)]) models"
        " concurrency with the sequential dependency A -> B. Avoid using a partial"
        " order as a child of"
        " another partial order to ensure not leaving out any sequential dependencies. To resolve this,"
        " you can combine the two orders.\n"
        " - decision_graph(dependencies) takes 1 argument, which is a list of tuples of submodels. These tuples"
        " define the nodes and the directed edges of a decision graph, which models control flow.\n\n"
        " A decision graph is a directed graph with implicitly defined start and end points."
        " You can define start and end nodes by using None as a source or target in the tuples. For example, (None, A): A is a start node; (B, None): B is an end node. "
        " Each tuple in the `dependencies` list represents a sequential pair"
        " For example, a list of tuples `[(A, B), (A, C)]` will create the directed edges"
        " `A -> B` and `A -> C`."
        " This structure is used to model complex process flows, where exactly one path is taken at a time. This is the main difference"
        " from a partial order: decision graphs model exclusive paths, i.e., structures"
        " where exactly one path can be executed. Partial orders, on the other hand, are used to model concurrency."
        " Assume we have 4 submodels A, B, C, D."
        " - `decision_graph(dependencies=[(None, A), (A, B), (A, C), (B, D), (B, C), (C, None), (D, None)])` models the following four sequences (paths):"
        " A->C, A->B->C, A->B->D where exactly one of these will be executed."
        " A is the only start node and C and D are the possible end nodes."
        " Decision graphs can also model cyclic behavior. For example, "
        " `decision_graph(dependencies=[(None, A), (A, B), (B, A), (A, None)])` models a loop between A and B,"
        " where A is the start and end node."
        " Decision graphs can also model exclusive choice structures of n >= 2 sub-models. For example, "
        " `decision_graph(dependencies=[(None, A), (None, B), (A, None), (B, None)])` models an exclusive choice between A and B."
        "IMPORTANT: a decision graph can additionally express simple sequential dependencies, e.g., A is followed by B."
        "Therefore, there is no need to add additional partial order to express them."
        "IMPORTANT: we define three helper functions:\n"
        " - copy(model) for any powl model 'model', you can always call it to get a copy"
        " of the same model. This is useful to model cases where a subporcess or activity can be executed exactly"
        "twice (not really in a loop)."
        " - skip(submodel) takes 1 argument, which is a submodel. It makes the submodel optional/skippable.\n"
        " - self_loop(submodel) takes 1 argument, which is a submodel. It creates a self-loop on the submodel, i.e., the subprocess must be executed at least once but can be repeated multiple times afterwards.\n"
        "IMPORTANT: You can model skippable loops by combining these two functions. For example, "
        "`skip(self_loop(A))` models a skippable loop of A, where A can be executed any number of times, including zero."
    )
    return prompt


def add_knowledge_about_resources():
    return (
        "Additionally, consider the following knowledge about pools and lanes:\n"
        "An activity may be assigned to a pool and a lane. A pool is used to define the boundaries of that participant's process. It typically refers to an organization."
        "A lane is used to define a distinct role within a pool."
        "Lanes help to organize and categorize activities within a pool based on the roles or responsibilities of different participants."
        """**Important** Do not generate too specific pool names, use generic ones like "Customer", "Supplier", "System", "Department", etc.\n"""
        "The name of each pool and lane should be derived from the process description."
        "Avoid repeating the same pool name for different pools.\n"
        "Make sure that each lane belongs to only one pool.\n"
        "Different **departments or roles within the same organization** should be modeled as lanes within one single pool.\n"
        "Sometimes, pools and lanes cannot be identified from the process description. In that case, you can assign them to 'None'.\n"
        "**Important** If you have managed to identify at least one pool, you cannot use 'None' for other pools.\n"
        "This is valid for lanes as well.\n"
        "IMPORTANT: DO NOT assign names to pools and lanes using variables, e.g., university_pool = 'University'. Always use string literals IN THE FUNCTION CALLS, e.g., pool='University'.\n"
    )


def add_process_description(process_description):
    return "This is the process description: " + process_description


def negative_prompting():
    return (
        "Avoid common mistakes. "
        "First, ensure that the transitive closure of the generated partial orders"
        " do not violate irreflexivity. Verify that all optional/skippable and"
        " repeatable parts are modeled correctly. Also validate that the same submodel"
        " is not used multiple times (e.g., in dependency_graph then in partial_oder)! You have three ways for avoiding"
        " this depending on the case: (1)"
        " consider using decision_graphs to model cyclic behaviour; (2) if you instead want to create a second instance"
        " of the same submodel, consider creating a copy of it; (3) if none of these two cases apply, then"
        " your structure is not correct. Ensure that you correctly model loops and exclusive choices between larger complete"
        " alternative/loop paths (i.e., between full paths, not decision points). Finally, do not create partial"
        " orders as children of other partial orders. "
        " Instead, combine dependencies at the same hierarchical level to avoid nested partial orders."
        " Example of Correct Use of Partial Order:\n"
        "```python\n"
        "poset = partial_order(dependencies=[(A, B), (B, C)])\n"
        "```\n\n"
        "Example of Incorrect Use of Partial Order:\n"
        "```python\n"
        "poset_1 partial_order(dependencies=[(B, C)])\n"
        "poset_2 = partial_order(dependencies=[(A, poset)])\n"
        "```\n\n"
        " A decision graph can have loops:"
        " a cycle (e.g., `[(None, A), (A, B), (B, A), (A, None)]`) means that you always start with A, then you can go to B, then back to A, and so on."
        "However, the cycle must include at least one start node and one end node (can be the same node)."
        "When using decision graphs, **keep the model as flat as possible**"
        "Examples of unflattened model:"
        "```python\n"
        "activity_block = partial_order(dependencies=[(A, B)])\n"
        "decision_graph = `decision_graph(dependencies=[(None, C), (C, D), (C, E), (E, None)])`\n"
        "model = partial_order(dependencies(activity_block, decision_graph))`\n"
        "```\n\n"
        "Instead, do this:"
        "```python\n"
        "model = decision_graph(dependencies=[(None, A), (A, B), (B, C), (C, D), (C, E), (E, None)])\n"
        "```\n\n"
    )


def code_generation():
    return (
        "At the end of your response provide a single Python code snippet (i.e., staring with '```python') that"
        " contains the full final code. \n\n"
    )


def add_few_shots(resource_aware_discovery=False):
    res = (
        "Please use few-shots learning. These are few illustrating shots extended with common errors that you"
        " should avoid for each example:\n"
    )
    used_shots = RESOURCE_AWARE_SHOTS if resource_aware_discovery else SHOTS
    for i in range(len(used_shots)):
        description, model, errors = used_shots[i]
        full_source = inspect.getsource(model)
        source_lines = full_source.split("\n")
        content_lines = source_lines[1:-2] + ["\n"]
        content_as_string = "\n".join(line[4:] for line in content_lines)
        res = res + f"Process description for example {i + 1}:\n{description}\n"
        res = res + f"Process model for example {i + 1}:\n"
        res = res + f"```python\n{import_statement}\n{content_as_string}"
        res = res + f"```\n"
        res = res + f"Common errors to avoid for example {i + 1}:\n{errors}\n"
    return res + "\n"


def create_model_generation_prompt(
    process_description: str, resource_aware_discovery: bool
) -> str:
    prompt = add_role()
    if resource_aware_discovery:
        prompt = prompt + add_knowledge_about_resources()
    prompt = prompt + add_knowledge(resource_aware_discovery)
    prompt = prompt + negative_prompting()
    prompt = prompt + add_few_shots(resource_aware_discovery)
    prompt = prompt + code_generation()

    if process_description is not None:
        prompt = prompt + add_process_description(process_description)

    return prompt


def create_conversation(
    process_description: Optional[str], resource_aware_discovery: bool
) -> List[dict[str:str]]:
    prompt = create_model_generation_prompt(
        process_description, resource_aware_discovery=resource_aware_discovery
    )
    conversation = [{"role": "user", "content": f"{prompt}"}]
    return conversation


def update_conversation(
    conversation: List[dict[str:str]], feedback: str
) -> List[dict[str:str]]:
    update_prompt = (
        "Please update the model to fix it based on the provided feedback. Please make sure the returned"
        " model matches the initial process description, all previously provided feedback, and the new"
        "feedback comment as well. Make sure to save the updated final model is the variable"
        " 'final_model'. This is the new feedback comment:" + feedback
    )
    conversation.append({"role": "user", "content": f"{update_prompt}"})
    return conversation


def cut_conversation(
    conversation: List[dict[str:str]], pos: int
) -> List[dict[str:str]]:

    if pos < 0 or pos >= len(conversation):
        raise ValueError("Invalid position to cut the conversation!")

    cut_convo = conversation[: pos + 1]
    # Ensure the last message is from the user and is not feedback
    while cut_convo:
        if (
            cut_convo[-1]["role"] != "user"
            or cut_convo[-1].get("type", "msg") == "error"
        ):
            cut_convo.pop()
        else:
            break
    if not cut_convo:
        raise Exception("No valid user message found to cut the conversation!")
    return cut_convo


def model_self_improvement_prompt():
    return (
        "Thank you! The model was generated successfully! Could you further improve the model? "
        "Please critically evaluate the process model and improve it accordingly **only where genuinely beneficial**. "
        "Potential improvement steps might for instance include adding missing activities, managing additional "
        "exceptions, increasing concurrency in execution, or elevating choices to higher levels. If you find the "
        "model already optimized or see no significant areas for enhancement, it is "
        "perfectly acceptable to make minimal adjustments (e.g., relabeling some activities) or to return the "
        "same model without any changes."
    )


def model_self_improvement_prompt_short():
    return (
        "Thank you! The model was generated successfully! Could you further improve the model? "
        "Please critically evaluate the process model against the initial process description and improve it"
        " accordingly **only where genuinely beneficial**. If you see no significant areas for enhancement, it is "
        "perfectly acceptable to return the same model without any changes. Regardless of whether you improve the"
        " model or not, make sure to include a Python snippet that contains the model in your response."
    )


def description_self_improvement_prompt(descr: str):
    res = f"""
    You are provided with a process description. Your task is to optimize this description to make it richer and more
     detailed, while ensuring that all additions are relevant, accurate, and directly related to the original process.
     The goal is to make the description more comprehensive and suitable for process modeling purposes. """

    res = (
        res
        + f"""Possible areas for enhancement include:\n
    - **Detail Enhancement:** Add specific details that are missing but crucial for understanding the process flow. \n
    - **Clarity Improvement:** Clarify any ambiguous or vague statements to ensure that the description is clear and understandable.\n
    - **Explicit Process Constructs:** Rephrase parts of the description to explicitly incorporate constructs.
    For example, change 'X happens in most cases' to 'there is an exclusive choice between performing X or
     skipping it'.\n

    Please answer by only returning the improved process description without any additional text in your response. Do
     not define concrete activity labels yourself!"\n

    The process description:\n

    {descr}
    """
    )

    return res



from copy import copy
from itertools import combinations

from pm4py.objects.petri_net.obj import PetriNet
from pm4py.objects.petri_net.utils import petri_utils as pn_util

from promoai.pn_to_powl.converter_utils.subnet_creation import (
    add_arc_from_to,
    clone_place,
    pn_transition_to_powl,
)
from promoai.pn_to_powl.converter_utils.weak_reachability import (
    get_reachable_transitions_from_place_to_another,
)


def mine_base_case(net: PetriNet):
    if len(net.transitions) == 1 and len(net.places) == 2 == len(net.arcs):
        activity = list(net.transitions)[0]
        powl_transition = pn_transition_to_powl(activity)
        return powl_transition
    return None


def mine_self_loop(
    net: PetriNet, start_place: PetriNet.Place, end_place: PetriNet.Place
):
    if start_place == end_place:
        place = start_place
        place_copy = clone_place(net, place, {})
        redo = copy(net.transitions)
        out_arcs = place.out_arcs
        for arc in list(out_arcs):
            target = arc.target
            pn_util.remove_arc(net, arc)
            add_arc_from_to(place_copy, target, net)
        do_transition = PetriNet.Transition(f"silent_do_{place.name}", None)
        do = set()
        do.add(do_transition)
        net.transitions.add(do_transition)
        add_arc_from_to(place, do_transition, net)
        add_arc_from_to(do_transition, place_copy, net)
        return do, redo, place, place_copy

    return None


def mine_loop(net: PetriNet, start_place: PetriNet.Place, end_place: PetriNet.Place):
    redo_subnet_transitions = get_reachable_transitions_from_place_to_another(
        end_place, start_place
    )

    if len(redo_subnet_transitions) == 0:
        return None, None

    do_subnet_transitions = get_reachable_transitions_from_place_to_another(
        start_place, end_place
    )

    if len(do_subnet_transitions) == 0:
        raise Exception("This should not be possible!")

    if do_subnet_transitions & redo_subnet_transitions:
        # This could happen if we have ->(..., Loop)
        return None, None

    if net.transitions != (do_subnet_transitions | redo_subnet_transitions):
        raise Exception("Something went wrong!")

    # A loop is detected: the set of transitions is partitioned into two disjoint, non-empty subsets (do and redo)
    return do_subnet_transitions, redo_subnet_transitions


def mine_xor(net: PetriNet, reachability_map):
    choice_branches = [{t} for t in net.transitions]

    for t1, t2 in combinations(net.transitions, 2):
        if t1 in reachability_map[t2] or t2 in reachability_map[t1]:
            new_branch = {t1, t2}
            choice_branches = __combine_parts(new_branch, choice_branches)

    if net.transitions != set().union(*choice_branches):
        raise Exception("This should not happen!")

    return choice_branches


def mine_partial_order(net, end_place, reachability_map):
    partition = [{t} for t in net.transitions]

    for place in net.places:
        out_size = len(place.out_arcs)
        if out_size > 1 or (place == end_place and out_size > 0):
            xor_branches = []
            for start_transition in pn_util.post_set(place):
                new_branch = {
                    node
                    for node in reachability_map[start_transition]
                    if isinstance(node, PetriNet.Transition)
                }
                xor_branches.append(new_branch)
            union_of_branches = set().union(*xor_branches)
            if place == end_place:
                not_in_every_branch = union_of_branches
            else:
                intersection_of_branches = set.intersection(*xor_branches)
                not_in_every_branch = union_of_branches - intersection_of_branches
            if len(not_in_every_branch) > 1:
                partition = __combine_parts(not_in_every_branch, partition)

    return partition


def __combine_parts(
    transitions_to_group_together: set[PetriNet.Transition],
    partition: list[set[PetriNet.Transition]],
):
    new_partition = []
    new_combined_group = set()

    for part in partition:

        if part & transitions_to_group_together:
            new_combined_group.update(part)
        else:
            new_partition.append(part)

    if new_combined_group:
        new_partition.append(new_combined_group)

    return new_partition



from copy import deepcopy

import networkx as nx

import powl
from powl.objects.BinaryRelation import BinaryRelation
from powl.objects.obj import (
    DecisionGraph,
    Operator,
    OperatorPOWL,
    POWL,
    SilentTransition,
    StrictPartialOrder,
    Transition,
)


def get_node_type(node):
    if node.__class__ is powl.objects.obj.Transition:
        return f"Activity ({node.label})"
    elif node.__class__ is powl.objects.obj.StrictPartialOrder:
        return "PartialOrder"
    elif node.__class__ is powl.objects.obj.OperatorPOWL:
        if node.operator is powl.objects.obj.Operator.XOR:
            return "XOR"
        elif node.operator is powl.objects.obj.Operator.LOOP:
            return "LOOP"
        else:
            return node.operator.value
    elif node.__class__ is powl.objects.obj.DecisionGraph:
        return "DecisionGraph"
    else:
        return node.__class__


class ModelGenerator:
    def __init__(
        self,
        enable_nested_partial_orders=True,
        copy_duplicates=False,
        enable_nested_decision_graphs=True,
    ):
        self.used_as_submodel = []
        self.nested_partial_orders = enable_nested_partial_orders
        self.copy_duplicates = copy_duplicates
        self.nested_decision_graphs = enable_nested_decision_graphs

    def activity(self, label, pool: str = None, lane: str = None):
        return Transition(label, organization=pool, role=lane)

    def silent_transition(self):
        return SilentTransition()

    def create_model(self, node: POWL):
        if node is None:
            res = SilentTransition()
        else:
            if isinstance(node, str):
                node = self.activity(node)
            elif not isinstance(node, POWL):
                raise Exception(
                    f"Only POWL models are accepted as submodels! You provide instead: {type(node)}."
                )
            if node in self.used_as_submodel:
                if self.copy_duplicates:
                    res = deepcopy(node)
                else:
                    node_type = get_node_type(node)
                    raise Exception(
                        f"Ensure that"
                        f" each submodel is used uniquely! Avoid trying to"
                        f" reuse submodels that were used as children of other constructs (xor, loop,"
                        f" or partial_order) before! The error occured when trying to reuse a node of type {node_type}."
                    )
            else:
                res = node
        self.used_as_submodel.append(res)
        return res

    def xor(self, *args):
        if len(args) < 2:
            raise Exception("Cannot create an xor of less than 2 submodels!")
        children = [self.create_model(child) for child in args]
        res = OperatorPOWL(Operator.XOR, children)
        return res

    def loop(self, do, redo):
        if do is None and redo is None:
            raise Exception(
                "Cannot create an empty loop with both the do and redo parts missing!"
            )
        children = [self.create_model(do), self.create_model(redo)]
        res = OperatorPOWL(Operator.LOOP, children)
        return res

    def partial_order(self, dependencies):
        list_children = []
        for dep in dependencies:
            if isinstance(dep, tuple):
                for n in dep:
                    if n not in list_children:
                        list_children.append(n)
            elif isinstance(dep, POWL):
                if dep not in list_children:
                    list_children.append(dep)
            else:
                raise Exception(
                    "Invalid dependencies for the partial order! You should provide a list that contains"
                    " tuples of POWL models!"
                )
        if len(list_children) == 1:
            return list_children[0]
        if len(list_children) == 0:
            raise Exception("Cannot create a partial order over 0 submodels!")
        children = dict()
        for child in list_children:
            new_child = self.create_model(child)
            children[child] = new_child

        if self.nested_partial_orders:
            pass
        else:
            for child in children:
                if isinstance(child, StrictPartialOrder):
                    raise Exception(
                        "Do not use partial orders as 'direct children' of other partial orders."
                        " Instead, combine dependencies at the same hierarchical level. Note that it is"
                        " CORRECT to have 'partial_order > xor/loop > partial_order' in the hierarchy,"
                        " while it is"
                        " INCORRECT to have 'partial_order > partial_order' in the hierarchy.'"
                    )
        order = StrictPartialOrder(list(children.values()))
        for dep in dependencies:
            if isinstance(dep, tuple):
                for i in range(len(dep) - 1):
                    source = dep[i]
                    target = dep[i + 1]
                    if source in children.keys() and target in children.keys():
                        order.add_edge(children[source], children[target])

        res = order
        return res

    def decision_graph(self, dependencies):
        list_children = []
        for dep in dependencies:
            if isinstance(dep, tuple):
                if len(dep) != 2:
                    raise Exception(
                        "Invalid dependency tuple in decision graph! Each tuple must contain exactly 2 elements."
                    )
                for n in dep:
                    if n not in list_children and n is not None:
                        list_children.append(n)
            elif isinstance(dep, POWL):
                if dep not in list_children:
                    list_children.append(dep)
            else:
                raise Exception(
                    "Invalid dependencies for the decision graph! You should provide a list that contains"
                    " tuples of POWL models!"
                )
        if len(list_children) < 1:
            raise Exception(
                ""
                "A decision graph has at least one node. The provided list should comprise of at least one element."
            )

        children = dict()
        for child in list_children:
            new_child = self.create_model(child)
            children[child] = new_child
        # Identify start and end nodes first
        # Default dict with values 0
        start_nodes, end_nodes = [], []
        empty_path = False

        for dep in dependencies:
            if isinstance(dep, tuple):
                # Add both of them there
                if len(dep) != 2:
                    raise Exception(
                        "Invalid dependency tuple in decision graph! Each tuple must contain exactly 2 elements. "
                        "Note that None can be used to indicate start or end."
                    )
                source = dep[0]
                target = dep[1]
                if source is None and target is None:
                    empty_path = True
                elif source is None:
                    start_nodes.append(children[target])
                elif target is None:
                    end_nodes.append(children[source])

        # Build a graph to check for edges
        G = nx.DiGraph()
        # Add artificial start and end nodes
        G.add_node("ArtificialStart")
        G.add_node("ArtificialEnd")
        for sn in start_nodes:
            G.add_edge("ArtificialStart", sn)
        for en in end_nodes:
            G.add_edge(en, "ArtificialEnd")
        if empty_path:
            G.add_edge("ArtificialStart", "ArtificialEnd")
        # We have everything needed to create the decision graph
        binary_relation = BinaryRelation(list_children)
        for dep in dependencies:
            if isinstance(dep, tuple):
                for i in range(len(dep) - 1):
                    source = dep[i]
                    target = dep[i + 1]
                    if source in children.keys() and target in children.keys():
                        G.add_edge(source, target)
                        binary_relation.add_edge(children[source], children[target])
        # Check if all nodes are from start to end
        for node in list_children:
            if not (
                nx.has_path(G, "ArtificialStart", node)
                and nx.has_path(G, node, "ArtificialEnd")
            ):
                raise Exception(
                    f"All nodes in a decision graph must be on a path from a start node to an end node, {children[node]} isn't!"
                )
        order = DecisionGraph(
            binary_relation, start_nodes, end_nodes, empty_path=empty_path
        )
        children = order.children
        if self.nested_decision_graphs:
            pass
        else:
            for child in children:
                if isinstance(child, DecisionGraph):
                    raise Exception(
                        "Do not use decision graphs as 'direct children' of other decision graphs."
                        " Instead, combine dependencies at the same hierarchical level in the same structure. Note that it is"
                        " CORRECT to have 'decision_graph > xor/loop/partial_order > decision_graph in the hierarchy"
                    )
        return order

    def self_loop(self, node: POWL):
        if node is None:
            raise Exception("Cannot create a self-loop over an empty model!")
        child = self.create_model(node)
        silent = SilentTransition()
        return OperatorPOWL(Operator.LOOP, [child, silent])

    def skip(self, node: POWL):
        if node is None:
            raise Exception("Cannot create a skip over an empty model!")
        child = self.create_model(node)
        silent = SilentTransition()
        return OperatorPOWL(Operator.XOR, [child, silent])

    @staticmethod
    def copy(node: POWL):
        return deepcopy(node)



from enum import Enum


class AIProviders(Enum):
    GOOGLE = "Google"
    OPENAI = "OpenAI"
    DEEPSEEK = "DeepSeek"
    ANTHROPIC = "Anthropic"
    DEEPINFRA = "Deepinfra"
    MISTRAL_AI = "Mistral AI"
    OPENROUTER = "OpenRouter"
    COHERE = "Cohere"
    GROK = "Grok"


AI_MODEL_DEFAULTS = {
    AIProviders.GOOGLE.value: "gemini-3-flash",
    AIProviders.OPENAI.value: "gpt-5.2-codex",
    AIProviders.DEEPSEEK.value: "deepseek-reasoner",
    AIProviders.ANTHROPIC.value: "claude-sonnet-4-20250514",
    AIProviders.DEEPINFRA.value: "openai/gpt-oss-20b",
    AIProviders.MISTRAL_AI.value: "mistral-large-latest",
    AIProviders.OPENROUTER.value: "openai/gpt-oss-20b",
    AIProviders.COHERE.value: "command-r-plus",
    AIProviders.GROK.value: "grok-3",
}

DEFAULT_AI_PROVIDER = AIProviders.GOOGLE.value

AI_HELP_DEFAULTS = {
    AIProviders.GOOGLE.value: "Enter a Google model name. You can get a **free Google API key** and check the latest models under: https://ai.google.dev/.",
    AIProviders.OPENAI.value: "Enter an OpenAI model name. You can get an OpenAI API key and check the latest models under: https://openai.com/pricing.",
    AIProviders.DEEPSEEK.value: "Enter a DeepSeek model name. You can get a DeepSeek API key and check the latest models under: https://api-docs.deepseek.com/.",
    AIProviders.ANTHROPIC.value: "Enter an Anthropic model name. You can get an Anthropic API key and check the latest models under: https://www.anthropic.com/api.",
    AIProviders.DEEPINFRA.value: "Enter a model name available through Deepinfra. DeepInfra supports popular open-source large language models like Meta's LLaMa and Mistral, and it also enables custom model deployment. You can get a Deepinfra API key and check the latest models under: https://deepinfra.com/models.",
    AIProviders.MISTRAL_AI.value: "Enter a Mistral AI model name. You can get a Mistral AI API key and check the latest models under: https://mistral.ai/.",
    AIProviders.OPENROUTER.value: "Enter an OpenRouter model name. You can get an OpenRouter API key and check the latest models under: https://openrouter.ai/models. Note that free models are available, but you may need to sign up for an account.",
    AIProviders.COHERE.value: "Enter a Cohere model name. You can get a Cohere API key and check the latest models under: https://cohere.com/docs/models. They further provide free trial keys.",
    AIProviders.GROK.value: "Enter a Grok model name. You can get a Grok API key and check the latest models under: https://x.ai/api.",
}

MAIN_HELP = (
    "Select the AI provider you'd like to use. Google offers the Gemini models,"
    " which you can **try for free**,"
    " while OpenAI provides GPT models. DeepInfra supports popular open-source large language models like"
    " Meta's LLaMa and also enables custom model deployment. You can also try the models provided by"
    " DeepSeek, Anthropic, Mistral AI, Grok, and Cohere."
    " OpenRouter is a platform that allows you to access various models from different providers."
    " For some of them, you may have free access, but you may need to sign up for an account."
)



from promoai.aipa.abstraction import get_simplified_xml_abstraction
from promoai.aipa.prompting import add_prompt_strategies


def create_conversation(
    role="system",
    model_abstraction="simplified_xml",
    enable_role_prompting=True,
    enable_natural_language_restriction=True,
    enable_chain_of_thought=False,
    enable_process_analysis=False,
    enable_knowledge_injection=False,
    enable_few_shots_learning=False,
    enable_negative_prompting=False,
    enable_examples=True,
):

    prompt = add_prompt_strategies(
        model_abstraction,
        enable_role_prompting,
        enable_natural_language_restriction,
        enable_chain_of_thought,
        enable_process_analysis,
        enable_knowledge_injection,
        enable_few_shots_learning,
        enable_negative_prompting,
        enable_examples,
    )
    conversation = [
        create_message(prompt, role=role, model_abstraction=model_abstraction)
    ]
    return conversation


def create_message(
    message: str,
    role,
    model_abstraction,
    additional_content=None,
    additional_content_type: str = "text",
):

    if model_abstraction not in ["svg"]:
        content = message
    else:
        content = [{"type": "text", "text": f"{message}"}]
        if additional_content is not None:
            content.append(
                {
                    "type": additional_content_type,
                    additional_content_type: additional_content,
                }
            )

    message = {"role": role, "content": content}

    return message


def create_process_model_representation(
    model_abstraction,
    role: str = "system",
    xml_string: str = None,
    json_abstraction: str = None,
    selected_elements_json: str = None,
) -> dict:

    if model_abstraction == "json":
        if json_abstraction is None:
            raise ValueError("json_abstraction must be provided")
        message = f"This is a JSON representing the full BPMN model: {json_abstraction}"

    elif model_abstraction == "xml":
        if xml_string is None:
            raise ValueError("xml_string must be provided")
        message = (
            f"This is a text containing the BPMN 2.0 XML of the process: {xml_string}"
        )

    elif model_abstraction == "svg":
        raise ValueError("SVG is not supported!")

    elif model_abstraction == "simplified_xml":
        if xml_string is None:
            raise ValueError("xml_string must be provided")
        simplified_xml_string = get_simplified_xml_abstraction(xml_string)
        message = f"This is an XML-like textual representation of the BPMN: {simplified_xml_string}"

    else:
        raise Exception(f"Unknown model abstraction: {model_abstraction}")

    if selected_elements_json:
        message = (
            message
            + f"\n \n The user has selected the following elements of the BPMN model (represented as a json): {selected_elements_json}"
        )

    abstraction_message = create_message(
        message, role=role, model_abstraction=model_abstraction
    )

    return abstraction_message



from promoai.general_utils.ai_providers import AIProviders
from promoai.general_utils.app_utils import InputType
from promoai.main import (
    generate_model_from_bpmn,
    generate_model_from_event_log,
    generate_model_from_petri_net,
    generate_model_from_text,
    query_bpmn,
)

__name__ = "promoai"
__version__ = "2.0.0"



from typing import Set, Union

from pm4py.objects.petri_net.obj import PetriNet
from pm4py.objects.petri_net.utils import petri_utils as pn_util
from powl.objects.obj import SilentTransition, Transition


def id_generator():
    count = 1
    while True:
        yield f"id{count}"
        count += 1


def clone_place(net, place, node_map):
    cloned_place = PetriNet.Place(f"{place.name}_cloned")
    net.places.add(cloned_place)
    node_map[place] = cloned_place
    return cloned_place


def clone_transition(net, transition, node_map):
    cloned_transition = PetriNet.Transition(
        f"{transition.name}_cloned", transition.label
    )
    net.transitions.add(cloned_transition)
    node_map[transition] = cloned_transition
    return cloned_transition


def clone_subnet(
    net: PetriNet,
    subnet_transitions: Set[PetriNet.Transition],
    start_place: PetriNet.Place,
    end_place: PetriNet.Place,
):
    subnet_net = PetriNet(f"Subnet_{next(id_generator())}")
    node_map = {}

    for node in subnet_transitions:
        clone_transition(subnet_net, node, node_map)

    # Add arcs and remaining places of the subnet
    for arc in net.arcs:
        source = arc.source
        target = arc.target
        if source in subnet_transitions or target in subnet_transitions:
            if source in node_map.keys():
                cloned_source = node_map[source]
            else:
                cloned_source = clone_place(subnet_net, source, node_map)
            if target in node_map.keys():
                cloned_target = node_map[target]
            else:
                cloned_target = clone_place(subnet_net, target, node_map)
            add_arc_from_to(cloned_source, cloned_target, subnet_net)

    mapped_start_place = node_map[start_place]
    mapped_end_place = node_map[end_place]

    return subnet_net, mapped_start_place, mapped_end_place


def locally_identical(p1, p2, transitions):
    pre1 = pn_util.pre_set(p1) & transitions
    pre2 = pn_util.pre_set(p2) & transitions
    post1 = pn_util.post_set(p1) & transitions
    post2 = pn_util.post_set(p2) & transitions
    return pre1 == pre2 and post1 == post2


def apply_partial_order_projection(
    net: PetriNet,
    subnet_transitions: Set[PetriNet.Transition],
    start_places: Set[PetriNet.Place],
    end_places: Set[PetriNet.Place],
):
    subnet_net = PetriNet(f"Subnet_{next(id_generator())}")
    node_map = {}

    for node in subnet_transitions:
        clone_transition(subnet_net, node, node_map)

    list_start_places = list(start_places)
    old_start = list_start_places[0]
    for place in list_start_places[1:]:
        if not locally_identical(place, old_start, subnet_transitions):
            raise Exception("Unique local start property is violated!")
    new_start_place = clone_place(subnet_net, old_start, node_map)
    node_map[old_start] = new_start_place

    if start_places == end_places:
        new_end_place = new_start_place
    else:
        list_end_places = list(end_places)
        old_end = list_end_places[0]
        for place in list_end_places[1:]:
            if not locally_identical(place, old_end, subnet_transitions):
                raise Exception("Unique local end property is violated!")
        new_end_place = clone_place(subnet_net, old_end, node_map)
        node_map[old_end] = new_end_place

    # Add arcs and remaining places of the subnet
    for arc in net.arcs:
        source = arc.source
        target = arc.target
        if source in subnet_transitions or target in subnet_transitions:
            if source in node_map.keys():
                cloned_source = node_map[source]
            else:
                if source in start_places or source in end_places:
                    continue
                cloned_source = clone_place(subnet_net, source, node_map)
            if target in node_map.keys():
                cloned_target = node_map[target]
            else:
                if target in start_places or target in end_places:
                    continue
                cloned_target = clone_place(subnet_net, target, node_map)
            add_arc_from_to(cloned_source, cloned_target, subnet_net)

    return subnet_net, new_start_place, new_end_place


def add_arc_from_to(
    source: Union[PetriNet.Place, PetriNet.Transition],
    target: Union[PetriNet.Transition, PetriNet.Place],
    net: PetriNet,
):
    arc = PetriNet.Arc(source, target)
    net.arcs.add(arc)
    source.out_arcs.add(arc)
    target.in_arcs.add(arc)


def pn_transition_to_powl(transition: PetriNet.Transition) -> Transition:
    label = transition.label
    if label:
        return Transition(label=label)
    else:
        return SilentTransition()
